Package: BiocSet
Title: Representing Different Biological Sets
Version: 1.24.0
Authors@R: c(
    person("Kayla", "Morrell", email = "kayla.morrell16@gmail.com",
        role = c("aut", "cre")),
    person("Martin", "Morgan", role = "aut"),
    person("Kevin", "Rue-Albrecht", role = "ctb"),
    person("Lluís", "Revilla Sancho", role = "ctb"))
Description: BiocSet displays different biological sets in a triple tibble
    format. These three tibbles are `element`, `set`, and `elementset`. The
    user has the abilty to activate one of these three tibbles to perform common
    functions from the dplyr package. Mapping functionality and accessing web
    references for elements/sets are also available in BiocSet.
Depends: R (>= 3.6), dplyr
Imports: methods, tibble, utils, rlang, plyr, S4Vectors, BiocIO,
        AnnotationDbi, KEGGREST, ontologyIndex, tidyr
Suggests: GSEABase, airway, org.Hs.eg.db, DESeq2, limma, BiocFileCache,
        GO.db, testthat, knitr, rmarkdown, BiocStyle
biocViews: GeneExpression, GO, KEGG, Software
License: Artistic-2.0
Encoding: UTF-8
RoxygenNote: 7.2.3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/BiocSet
git_branch: RELEASE_3_22
git_last_commit: f0c8b16
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 02:49:21 UTC; biocbuild
Author: Kayla Morrell [aut, cre],
  Martin Morgan [aut],
  Kevin Rue-Albrecht [ctb],
  Lluís Revilla Sancho [ctb]
Maintainer: Kayla Morrell <kayla.morrell16@gmail.com>

# Generated by roxygen2: do not edit by hand

S3method(arrange,BiocSet)
S3method(as.list,BiocSet)
S3method(es_activate,BiocSet)
S3method(filter,BiocSet)
S3method(group_by,BiocSet)
S3method(intersect,BiocSet)
S3method(left_join,BiocSet)
S3method(map_set,BiocSet)
S3method(mutate,BiocSet)
S3method(select,BiocSet)
S3method(summarise,BiocSet)
S3method(union,BiocSet)
export("metadata<-")
export(BiocSet)
export(BiocSet_from_GeneSetCollection)
export(BiocSet_from_elementset)
export(GeneSetCollection_from_BiocSet)
export(OBOSet)
export(arrange_element)
export(arrange_elementset)
export(arrange_set)
export(data.frame_from_element)
export(data.frame_from_elementset)
export(data.frame_from_set)
export(es_activate)
export(filter_element)
export(filter_elementset)
export(filter_set)
export(go_sets)
export(intersect_single)
export(kegg_sets)
export(left_join_element)
export(left_join_elementset)
export(left_join_set)
export(map_add_element)
export(map_add_set)
export(map_multiple)
export(map_set)
export(map_unique)
export(metadata)
export(mutate_element)
export(mutate_elementset)
export(mutate_set)
export(oboset_element_ancestors)
export(oboset_element_children)
export(oboset_element_parents)
export(oboset_set_ancestors)
export(oboset_set_children)
export(oboset_set_parents)
export(select_element)
export(select_elementset)
export(select_set)
export(summarise_element)
export(summarise_elementset)
export(summarise_set)
export(tibble_from_element)
export(tibble_from_elementset)
export(tibble_from_set)
export(union_single)
export(url_ref)
export(url_ref_element)
export(url_ref_set)
exportClasses(BiocSet)
exportClasses(OBOSet)
exportMethods(coerce)
exportMethods(es_element)
exportMethods(es_elementset)
exportMethods(es_set)
exportMethods(export)
exportMethods(import)
importClassesFrom(S4Vectors,Annotated)
importFrom(AnnotationDbi,keys)
importFrom(AnnotationDbi,keytypes)
importFrom(AnnotationDbi,mapIds)
importFrom(BiocIO,export)
importFrom(BiocIO,import)
importFrom(BiocIO,resource)
importFrom(KEGGREST,keggGet)
importFrom(KEGGREST,keggList)
importFrom(S4Vectors,"metadata<-")
importFrom(S4Vectors,metadata)
importFrom(dplyr,'%>%')
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,intersect)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_if)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,right_join)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,summarise_all)
importFrom(dplyr,tbl_nongroup_vars)
importFrom(dplyr,ungroup)
importFrom(dplyr,union)
importFrom(methods,coerce)
importFrom(methods,initialize)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setAs)
importFrom(methods,setOldClass)
importFrom(methods,slot)
importFrom(ontologyIndex,get_ontology)
importFrom(plyr,mapvalues)
importFrom(rlang,.data)
importFrom(rlang,enquo)
importFrom(rlang,quo_text)
importFrom(stats,setNames)
importFrom(tibble,as_tibble)
importFrom(tibble,enframe)
importFrom(tibble,is_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,unnest)
importFrom(utils,write.table)

#' @importFrom AnnotationDbi mapIds keytypes
#' @importFrom tibble enframe
.es_map <- function(es, org, from, to, multi)
{
    stopifnot(
        all(es_element(es)$element %in%
            keys(org, keytype = from)),
        from %in% keytypes(org),
        to %in% keytypes(org)
    )

    keys <- es_element(es)$element
    map <- mapIds(org, keys, to, from, multiVals = multi)
    if (any(is(map) == "SimpleCharacterList"))
        map <- as.vector(map)
    tbl <- enframe(map, name = from, value = to)
    es %>% map_element(tbl[[from]], tbl[[to]])
}

.normalize_mapping <-
    function(from, to)
{
    stopifnot(
        length(from) == length(to),
        is.null(names(from)), is.null(names(to))
    )

    to <- rep(to, lengths(from))
    from <- unlist(from)

    from <- rep(from, lengths(to))
    to <- unlist(to)

    tibble(element = unname(from), to = unname(to))
}

map_element <- function(.data, from, to, keep_unmapped) UseMethod("map_element")

#' @importFrom tibble as_tibble
#' @importFrom dplyr bind_rows summarise_all mutate_if
map_element.BiocSet <-
    function(.data, from, to, keep_unmapped = TRUE)
{
    stopifnot(is.character(from),
        is.character(to) || is.list(to) || is(to, "CharacterList"),
        length(from) == length(to)
    )
    element <- set <- NULL

    es <- es_elementset(.data)

    ## mapping <- tibble(element = from, to)
    mapping <- .normalize_mapping(from, to)
    if (keep_unmapped) {
        aux <- as_tibble(es) %>%        # un-mapped elements
            select(element) %>%
            mutate(to = element) %>%
            filter(!element %in% from)
        mapping <- bind_rows(mapping, aux)
    }

    es <-                               # map
        left_join(mapping, es) %>%
        select(-element, element = to)
    es <- es %>%                        # de-duplicate
        group_by(element, set) %>%
        summarise_all(list) %>%
        mutate_if(.test, unlist)

    sets <- es_set(.data) %>%
        filter(set %in% es$set)

    elements <- es_element(.data)
    elements <-                         # map
        left_join(mapping, elements) %>%
        select(-element, element = to)
    elements <- elements %>%
        group_by(element) %>%
        summarise_all(list) %>%
        mutate_if(.test, unlist)

    BiocSet_from_elementset(es, elements, sets, metadata(.data))
}

#' Functions for mapping elements in the element tibble to different id types
#' @rdname mapping_element
#' @name mapping_element
#' @description Functions for dealing with unique mapping and multiple mapping.
#'     \code{map_add_element} will add the mapping as a new column instead of 
#'     overwriting the current one used for the mapping.  
#' @param es The BiocSet objec to map the elements on.
#' @param org The AnnotationDbi object to identify keys/mappings from.
#' @param from A character to indicate which identifier to map from.
#' @param to A character to indicate which identifier to map to.
#' @return For \code{map_unique}, a \code{BiocSet} object with unique 
#'    elements.
#' @export
#' @examples
#' library(org.Hs.eg.db)
#' es <- BiocSet(set1 = c("C5", "GANC"), set2 = c("AFM", "CGB1", "ADAM32"))
#' map_unique(es, org.Hs.eg.db, "SYMBOL", "ENTREZID")
map_unique <- function(es, org, from, to)
    .es_map(es, org, from, to, multi = "first")


#' @rdname mapping_element
#' @name mapping_element
#' @param multi How should multiple values be returned? 
#'     Options include:     
#'     \itemize{
#'     \item{list: This will just return a list object to the end user.}
#'     \item{filter: This will remove all elements that contain multiple 
#'     matches and will therefore return a shorter vector than what came in 
#'     whenever some of the keys match more than one value.}
#'     \item{asNA: This will return an NA value whenever there are multiple 
#'     matches.}
#'     \item{CharacterList: This just returns a SimpleCharacterList object.}
#'     \item{FUN: A function can be supplied to the 'multiVals' argument 
#'     for custom behaviors.}
#'     }
#' @export
#' @return For \code{map_multiple}, a \code{BiocSet} object with multiple 
#'     mappings for certain elements.
#' @examples
#' 
#' map_multiple(es, org.Hs.eg.db, "SYMBOL", "ENSEMBLTRANS", "asNA")
map_multiple <- function(es, org, from, to, multi =
    c('list', 'filter', 'asNA', 'CharacterList'))
{
    if(!is.function(multi))
        multi <- match.arg(multi)
    .es_map(es, org, from, to, multi)
}


#' @rdname mapping_element
#' @param add The id to add to the \code{BiocSet} object.
#' @return For \code{map_add_element}, a \code{BiocSet} object with a new column
#'     in the element tibble with the mapping of the new id type.
#' @export
#' @examples
#'
#' map <- map_add_element(es, org.Hs.eg.db, "SYMBOL", "ENTREZID")
#' es %>% mutate_element(entrez = map)
map_add_element <- function(es, org, from, add)
{
    stopifnot(from %in% keytypes(org),
        add %in% keytypes(org))

    map <- mapIds(org,
        keys = es_element(es)$element,
        column = add,
        keytype = from,
        multivals = "first"
    )
    unname(map)
}

context("mapping_element")

library(org.Hs.eg.db)

test_that("'.es_map()' works",
{
    metadata <- list(organism = "Homo sapiens")
    es <- BiocSet(
        set1 = c("BRCA1", "BRCA2", "TGFA", "ERCC2"),
        set2 = c("PRNP", "FMR1", "PAX3"),
        metadata = metadata
    )
    es1 <- es %>% .es_map(org.Hs.eg.db, "SYMBOL", "ENTREZID")

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(7L, 1L))
    expect_identical(dim(es_set(es1)), c(2L, 1L))
    expect_identical(dim(es_elementset(es1)), c(7L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(metadata(es1), metadata)

    expect_true(all(es_element(es1)$element %in%
        keys(org.Hs.eg.db, keytype="ENTREZID")))

    expect_error(.es_map(es, org.Hs.eg.db))
    expect_error(.es_map(es, org.Hs.eg.db, "SYMBOL"))
    expect_error(.es_map(es, org.Hs.eg.db, SYMBOL, "ENTREZID"))
    expect_error(.es_map(es, org.Hs.eg.db, "SYMBOL", 10))
    expect_error(.es_map(es, org.Hs.eg.db, "SYMBOL", "IDS"))
    expect_error(.es_map(es, species, "SYMBOL", "ENTREZID"))
})

test_that("'.normalize_mapping()' works", {
    es <- .normalize_mapping(letters, LETTERS)
    expect_s3_class(es, "tbl_df")
    expect_false(.is_tbl_elementset(es))
    expect_identical(dim(es), c(26L, 2L))
    expect_true(is.character(es$element))
    expect_true(is.character(es$to))

    es1 <- .normalize_mapping(
        list(letters[1], letters[2:3], character()),
        LETTERS[1:3]
    )
    expect_s3_class(es1, "tbl_df")
    expect_false(.is_tbl_elementset(es1))
    expect_identical(dim(es1), c(3L, 2L))
    expect_true(is.character(es$element))
    expect_true(is.character(es$to))

    es2 <- .normalize_mapping(
        letters[1:3],
        list(LETTERS[1], LETTERS[2:3], character())
    )
    expect_s3_class(es2, "tbl_df")
    expect_false(.is_tbl_elementset(es2))
    expect_identical(dim(es2), c(3L, 2L))
    expect_true(is.character(es$element))
    expect_true(is.character(es$to))

    es3 <- .normalize_mapping(
        list(letters[1:2], letters[3], character()),
        list(LETTERS[1], LETTERS[2:3], character())
    )
    expect_s3_class(es3, "tbl_df")
    expect_false(.is_tbl_elementset(es3))
    expect_identical(dim(es3), c(4L, 2L))
    expect_true(is.character(es$element))
    expect_true(is.character(es$to))
})

test_that("'map_element.BiocSet()' works", {
    metadata <- list(organism = "Homo sapiens")
    es <- BiocSet(set1 = letters, set2 = LETTERS, metadata = metadata)

    # 1:1 mapping
    es1 <- es %>% map_element(letters, LETTERS)
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_element(es1)), c(26L, 1L))
    expect_identical(dim(es_set(es1)), c(2L, 1L))
    expect_identical(dim(es_elementset(es1)), c(52L,2L))
    expect_identical(metadata(es1), metadata)

    # A single set example with metadata in each tibble
    element <- tibble(element = letters[1:5], foo = 1:5)
    set <- tibble(set = "set1", bar = 1)
    elementset <- tibble(
        element = letters[1:5], set = rep("set1", 5), baz = 1:5
    )
    .data = BiocSet_from_elementset(elementset, element, set, metadata(es))

    from = c("a", "b", "b", "c", "d")
    to = c("A", "B", "C", "D", "D")
    # 1:1 mapping, 1:many mapping, many:1 mapping
    es2 <- map_element(.data, from, to)
    expect_s4_class(es2, "BiocSet")
    expect_identical(dim(es_element(es2)), c(5L, 2L))
    expect_identical(dim(es_set(es2)), c(1L, 2L))
    expect_identical(dim(es_elementset(es2)), c(5L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_true(is.list(es_element(es2)$foo))
    expect_true(is.double(es_set(es2)$bar))
    expect_true(is.list(es_elementset(es2)$baz))
    expect_identical(metadata(es2), metadata)

    # 1:1 mapping, 1:many mapping, many:1 mapping, dropping unmapped element(s)
    es3 <- map_element(.data, from, to, keep_unmapped = FALSE)
    expect_s4_class(es3, "BiocSet")
    expect_identical(dim(es_element(es3)), c(4L, 2L))
    expect_identical(dim(es_set(es3)), c(1L, 2L))
    expect_identical(dim(es_elementset(es3)), c(4L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es3)))
    expect_true(is.list(es_element(es3)$foo))
    expect_true(is.double(es_set(es3)$bar))
    expect_true(is.list(es_elementset(es3)$baz))
    expect_identical(metadata(es3), metadata)

    # multiple sets with metadata in element and elementset tibble
    elementset <- tibble(
        element = c("a", "b", "c"),
        set = c("set1", "set1", "set2")
    )
    .data = BiocSet_from_elementset(elementset, metadata = metadata(es)) %>%
        mutate_element(foo = 1:3) %>%
        mutate_elementset(bar = 1:3)

    # 1:1 mapping, but in different sets
    from <- c("b", "c")
    to <- c("D", "E")
    es4 <- map_element(.data, from, to)
    expect_s4_class(es4, "BiocSet")
    expect_identical(dim(es_element(es4)), c(3L, 2L))
    expect_identical(dim(es_set(es4)), c(2L, 1L))
    expect_identical(dim(es_elementset(es4)), c(3L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es4)))
    expect_true(is.integer(es_element(es4)$foo))
    expect_true(is.integer(es_elementset(es4)$bar))
    expect_identical(metadata(es4), metadata)

    # 1:1 mappping, in different sets, dropping unmapped element(s)
    es5 <- map_element(.data, from, to, keep_unmapped = FALSE)
    expect_s4_class(es5, "BiocSet")
    expect_identical(dim(es_element(es5)), c(2L, 2L))
    expect_identical(dim(es_set(es5)), c(2L, 1L))
    expect_identical(dim(es_elementset(es5)), c(2L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es5)))
    expect_true(is.integer(es_element(es5)$foo))
    expect_true(is.integer(es_elementset(es5)$bar))
    expect_identical(to, es_element(es5)$element)
    expect_identical(to, es_elementset(es5)$element)
    expect_identical(metadata(es5), metadata)

    # many:1 mapping, in different sets
    from <- c("b", "c")
    to <- c("D", "D")
    es6 <- map_element(.data, from, to)
    expect_s4_class(es6, "BiocSet")
    expect_identical(dim(es_element(es6)), c(2L, 2L))
    expect_identical(dim(es_set(es6)), c(2L, 1L))
    expect_identical(dim(es_elementset(es6)), c(3L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es6)))
    expect_true(is.list(es_element(es6)$foo))
    expect_true(is.integer(es_elementset(es6)$bar))
    expect_identical(metadata(es6), metadata)

    # many:1 mapping, in different sets, dropping unmapped element(s)
    es7 <- map_element(.data, from, to, keep_unmapped = FALSE)
    expect_s4_class(es7, "BiocSet")
    expect_identical(dim(es_element(es7)), c(1L, 2L))
    expect_identical(dim(es_set(es7)), c(2L, 1L))
    expect_identical(dim(es_elementset(es7)), c(2L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es7)))
    expect_true(is.list(es_element(es7)$foo))
    expect_true(is.integer(es_elementset(es7)$bar))
    expect_identical(to, es_elementset(es7)$element)
    expect_identical(metadata(es7), metadata)

    # 1:many mapping, in different sets
    from <- c("b", "b", "c", "c")
    to <- c("D", "E", "F", "G")
    es8 <- map_element(.data, from, to)
    expect_s4_class(es8, "BiocSet")
    expect_identical(dim(es_element(es8)), c(5L, 2L))
    expect_identical(dim(es_set(es8)), c(2L, 1L))
    expect_identical(dim(es_elementset(es8)), c(5L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es8)))
    expect_true(is.integer(es_element(es8)$foo))
    expect_true(is.integer(es_elementset(es8)$bar))
    expect_identical(metadata(es8), metadata)

    # 1:many mapping, in different sets, dropping unmapped element(s)
    es9 <- map_element(.data, from, to, keep_unmapped = FALSE)
    expect_s4_class(es9, "BiocSet")
    expect_identical(dim(es_element(es9)), c(4L, 2L))
    expect_identical(dim(es_set(es9)), c(2L, 1L))
    expect_identical(dim(es_elementset(es9)), c(4L, 3L))
    expect_true(.is_tbl_elementset(es_elementset(es9)))
    expect_true(is.integer(es_element(es9)$foo))
    expect_true(is.integer(es_elementset(es9)$bar))
    expect_identical(to, es_element(es9)$element)
    expect_identical(to, es_elementset(es9)$element)
    expect_identical(metadata(es9), metadata)

    expect_error(es %>% map_element())
})

test_that("'map_unique()' works",
{
    metadata <- list(organism = "Homo sapiens")
    es <- BiocSet(
        set1 = c("PRKACA", "TGFA", "MAP2K1"),
        set2 = c("CREB3", "FOS"),
        metadata = metadata
    )

    es1 <- es %>% map_unique(org.Hs.eg.db, "SYMBOL", "ENSEMBL")

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(5L, 1L))
    expect_identical(dim(es_set(es1)), c(2L, 1L))
    expect_identical(dim(es_elementset(es1)), c(5L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(metadata(es1), metadata)

    expect_error(map_unique(es, org.Hs.eg.db, "SYMBOL", "ENTREZID", "first"))
})

test_that("'map_multiple()' works",
{
    metadata <- list(organism = "Homo sapiens")
    es <- BiocSet(
        set1 = c("CFB", "DDR1"),
        set2 = c("CLIC1", "DEFB4A", "A2M"),
        metadata = metadata
    )

    es1 <- es %>% map_multiple(org.Hs.eg.db, "SYMBOL", "ENSEMBL", "list")

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(27L, 1L))
    expect_identical(dim(es_set(es1)), c(2L, 1L))
    expect_identical(dim(es_elementset(es1)), c(27L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_true(is.character(es_element(es1)$element))
    expect_true(is.character(es_elementset(es1)$element))
    expect_identical(metadata(es1), metadata)

    es2 <- es %>% map_multiple(org.Hs.eg.db, "SYMBOL", "ENSEMBL", "filter")

    expect_s4_class(es2, "BiocSet")
    expect_identical(dim(es_element(es2)), c(5L, 1L))
    expect_identical(dim(es_set(es2)), c(2L, 1L))
    expect_identical(dim(es_elementset(es2)), c(5L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_identical(metadata(es2), metadata)

    es3 <- es %>% map_multiple(org.Hs.eg.db, "SYMBOL", "ENSEMBL", "asNA")

    expect_s4_class(es3, "BiocSet")
    expect_identical(dim(es_element(es3)), c(2L, 1L))
    expect_identical(dim(es_set(es3)), c(2L, 1L))
    expect_identical(dim(es_elementset(es3)), c(3L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es3)))
    expect_false(all(is.na(es_element(es3)$element)))
    expect_false(all(is.na(es_elementset(es3)$element)))
    expect_identical(metadata(es3), metadata)

    es4 <- map_multiple(es, org.Hs.eg.db, "SYMBOL", "ENSEMBL", "CharacterList")

    expect_s4_class(es4, "BiocSet")
    expect_identical(dim(es_element(es4)), c(27L, 1L))
    expect_identical(dim(es_set(es4)), c(2L, 1L))
    expect_identical(dim(es_elementset(es4)), c(27L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es4)))
    expect_true(is.character(es_element(es4)$element))
    expect_true(is.character(es_elementset(es4)$element))
    expect_identical(metadata(es4), metadata)

    expect_error(map_multiple(es, org.Hs.eg.db, "SYMBOL", "ENSEMBL", "first"))
    expect_error(map_multiple(es, org.Hs.eg.db, "ENSEMBL", "SYMBOL", "list"))
    expect_error(map_multiple(es, org.Hs.eg.db, "SYMBOL", "ENSEMBL", filter()))
    expect_error(map_multiple(es, org.Hs.eg.db))
})

test_that("'map_add_element()' works",
{
    es <- BiocSet(set1 = c("PRKACA", "TGFA", "MAP2K1"), set2 = c("FOS", "BRCA1"))
    map <- map_add_element(es, org.Hs.eg.db, "SYMBOL", "ENTREZID")

    expect_identical(length(map), 5L)
    expect_identical(class(map), "character")
    expect_identical(map, c("5566", "7039", "5604", "2353", "672"))

    expect_error(map_add_element(org.Hs.eg.db, "SYMBOL", "ENTREZID"))
})

#' Set representation as an S3 class tibble
#' @rdname tblset
#' @name tblset
#' @param tbl_elementset An S3 \code{elementset} tibble.
#' @return An S3 \code{set} object in a tibble representation.
#' @importFrom tibble tibble
#' @importFrom dplyr distinct '%>%'
#' @keywords internal
#' @examples
#' tbl <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)
#' BiocSet:::.tbl_set(tbl)

.tbl_set <-
    function(tbl_elementset)
{
    stopifnot(
        .is_tbl_elementset(tbl_elementset)
    )

    tbl <- tibble(set = tbl_elementset$set) %>% distinct()
    subclass_tbl_elementset_base(tbl, "tbl_set")
}

#' Functions applied to elementsets in a \code{BiocSet} object
#' @rdname elementset_funs
#' @name elementset_funs
#' @description All of the major methods applied to a \code{BiocSet} object can
#'     be explicitly applied to the elementset tibble. These functions bypass 
#'     the need to use the \code{es_activate} function by indicating what 
#'     function should be used on the elementset tibble.
#' @param .data A \code{BiocSet} object.
#' @param ... Additional arguments passed to the function.
#' @return A \code{BiocSet} object.
#' @export
#' @examples
#' es <- BiocSet(set1 = letters, set2 = LETTERS)
#' filter_elementset(es, element == "a" | element == "A")
filter_elementset <- function(.data, ...) {
        act <- .active(.data)
        tbl <- es_activate(.data, "elementset") %>% filter(...)
        initialize(tbl, active = act)
}

#' @rdname elementset_funs
#' @export
#' @examples
#' 
#' es %>% select_elementset(element)
select_elementset <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "elementset") %>% select(...)
    initialize(tbl, active = act)
}

#' @rdname elementset_funs
#' @export
#' @examples
#' 
#' es %>% mutate_elementset(pval = rnorm(1:52))
mutate_elementset <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "elementset") %>% mutate(...)
    initialize(tbl, active = act)
}

#' @rdname elementset_funs
#' @export
#' @examples
#' 
#' es %>% summarise_elementset(n = n())
summarise_elementset <- function(.data, ...) {
    act <- .active(.data)
    es_activate(.data, "elementset") %>% summarise(...)
}

#' @rdname elementset_funs
#' @export
#' @examples
#' 
#' es %>% arrange_elementset(desc(element))
arrange_elementset <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "elementset") %>% arrange(...)
    initialize(tbl, active = act)
}

#' @rdname elementset_funs
#' @export
#' @examples
#' 
#' tbl <- tibble(x = 5:6, y = c("set1", "set2"))
#' es %>% left_join_elementset(tbl, by = c(set = "y"))
left_join_elementset <- function(.data, ...)
{
    tbl <- es_elementset(.data) %>% left_join(...)
    initialize(.data, elementset = tbl)
}

#' @rdname elementset_funs
#' @return For \code{tibble_from_elementset}, a tibble.
#' @export
#' @examples
#' 
#' tibble_from_elementset(es)
tibble_from_elementset <-
    function(.data)
{
    stopifnot(is(.data, "BiocSet"))
    es_elementset(.data) %>%
        left_join(es_set(.data)) %>%
        left_join(es_element(.data))
}

#' @rdname elementset_funs
#' @return For \code{data.frame_from_elementset}, a data.frame.
#' @export
#' @examples
#' 
#' data.frame_from_elementset(es)
data.frame_from_elementset <-
    function(.data)
{
    tbl <- tibble_from_elementset(.data)
    data.frame(tbl)
}

#' Functions to access reference urls for different identifiers
#' @rdname url_ref
#' @name url_ref
#' @param es A \code{BiocSet} object that the reference urls should be added to.
#' @return For \code{url_ref_element}, a \code{BiocSet} object with the url 
#'     column added to the element tibble.
#' @export
#' @examples
#' es <- BiocSet("GO:0000002" = c("TP53", "TNF"), "GO:0000003" = c("IL6"))
#' url_ref_element(es) 
url_ref_element <- function(es) 
{
    elements <- es_element(es)$element
    
    url <- ifelse(startsWith(elements, "ENSG"),
        "https://www.ensembl.org/Homo_sapiens/Gene/Summary?db=core;g=",
        "https://www.ncbi.nlm.nih.gov/gene/?term="
    )

    url <- paste0(url, elements)
    es %>% mutate_element(url = url)
}

#' @rdname url_ref
#' @name url_ref 
#' @return For \code{url_ref_set}, a \code{BiocSet} object with the url column 
#'     added to the set tibble.
#' @export
#' @examples
#' 
#' url_ref_set(es)
url_ref_set <- function(es)
{
    sets <- as.character(es_set(es)$set)    

    url <- ifelse(startsWith(sets, "GO"),
        "http://amigo.geneontology.org/amigo/medial_search?q=",
        "https://www.genome.jp/dbget-bin/www_bget?pathway:"
    )

    url <- paste0(url, sets)
    es %>% mutate_set(url = url)
}

#' @rdname url_ref
#' @name url_ref
#' @return For \code{url_ref}, a \code{BiocSet} object with the url column 
#'     added to both the element and set tibbles.
#' @export
#' @examples
#'
#' url_ref(es)
url_ref <- function(es)
{
    es %>% url_ref_element() %>% url_ref_set()
}

context("url_ref")

test_that("'url_ref_element()' works",
{
    es <- BiocSet(set1 = c("TP53", "TNF", "EGFR"), set2 = c("ENSG00000136244", 
        "ENSG00000112715"), set3 = c("291", "1890", "4205"))
    es1 <- url_ref_element(es)

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(8L, 2L))
    expect_identical(dim(es_set(es1)), c(3L, 1L))
    expect_identical(dim(es_elementset(es1)), c(8L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))

    expect_error(url_ref_element(1))
    expect_error(url_ref_element())
})

test_that("'url_ref_set()' works",
{
    es <- BiocSet("GO:0000002" = c("abc", "ABC"), "GO:0000010" = c("def", "ghi",
        "DEF", "GHI"), "hsa00010" = c("jkl", "JKL", "mno", "MNO"), "hsa00020" =
        c("pqr", "PQR"))
    es1 <- url_ref_set(es)

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(12L, 1L))
    expect_identical(dim(es_set(es1)), c(4L, 2L))
    expect_identical(dim(es_elementset(es1)), c(12L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))

    expect_error(url_ref_set(3))
    expect_error(url_ref_set())
})

test_that("'url_ref()' works",
{
    es <- BiocSet("GO:0000002" = c("abc", "ABC"), "GO:0000010" = c("def", "ghi",
        "DEF", "GHI"), "hsa00010" = c("jkl", "JKL", "mno", "MNO"), "hsa00020" = 
        c("pqr", "PQR"))
    es1 <- url_ref(es)

    expect_s4_class(es1, "BiocSet")
    expect_identical(dim(es_element(es1)), c(12L, 2L))
    expect_identical(dim(es_set(es1)), c(4L, 2L))
    expect_identical(dim(es_elementset(es1)), c(12L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es1)))

    expect_error(url_ref(3))
    expect_error(url_ref())
    expect_error(url_ref("abc"))
})

#' Functions applied to sets in a \code{BiocSet} object
#' @rdname set_funs
#' @name set_funs
#' @description All of the major methods applied to a \code{BiocSet} object can
#'     be explicitly applied to the set tibble. These functions bypass the need
#'     to use the \code{es_activate} function by indicating what function should
#'     be used on the element tibble.
#' @param .data A \code{BiocSet} object.
#' @param ... Additional argument passed to the function.
#' @return A \code{BiocSet} object.
#' @export
#' @examples
#' es <- BiocSet(set1 = letters, set2 = LETTERS)
#' filter_set(es, set == "set1")
filter_set <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "set") %>% filter(...)
    initialize(tbl, active = act)
}

#' @rdname set_funs
#' @name set_funs
#' @export
#' @examples
#' 
#' es %>% select_set(set)
select_set <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "set") %>% select(...)
    initialize(tbl, active = act)
}

#' @rdname set_funs
#' @name set_funs
#' @export
#' @examples
#' 
#' es %>% mutate_set(pval = rnorm(1:2))
mutate_set <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "set") %>% mutate(...)
    initialize(tbl, active = act)
}

#' @rdname set_funs
#' @name set_funs
#' @export
#' @examples
#' 
#' es %>% summarise_set(n = n())
summarise_set <- function(.data, ...) {
    act <- .active(.data)
    es_activate(.data, "set") %>% summarise(...)
}

#' @rdname set_funs
#' @name set_funs
#' @export
#' @examples
#' 
#' es %>% arrange_set(desc(set))
arrange_set <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "set") %>% arrange(...)
    initialize(tbl, active = act)
}

#' @rdname set_funs
#' @name set_funs
#' @export
#' @examples
#' 
#' tbl <- tibble(x = 10:11, y = c("set1", "set2"))
#' es <- BiocSet(set1 = letters[c(1,3,5)], set2 = letters[c(2,4)])
#' left_join_set(es, tbl, by = c(set = "y"))
left_join_set <- function(.data, ...)
{
    tbl <- es_set(.data) %>% left_join(...)
    initialize(.data, set = tbl)
}

#' @rdname set_funs
#' @name set_funs
#' @param how Multiple entries will become a list.
#' @return For \code{tibble_from_set}, a tibble.
#' @export
#' @examples
#' 
#' tibble_from_set(es)
tibble_from_set <-
    function(.data, how = unlist)
{
    set <- NULL
    tibble_from_elementset(.data) %>%
        group_by(set) %>%
        summarise_all(list) %>%
        mutate_if(.test, how)
}

#' @rdname set_funs
#' @name set_funs
#' @return For \code{data.frame_from_set}, a data.frame.
#' @export
#' @examples
#' 
#' data.frame_from_set(es)
data.frame_from_set <-
    function(.data, how = unlist)
{
    tbl <- tibble_from_set(.data, how)
    data.frame(tbl, row.names = "set")
}

#' Element set representation as an S3 class tibble
#' @rdname tblelementset
#' @name tblelementset
#' @param ... For \code{tbl_elementset}, named character vectors of gene sets. 
#'     Each character vector is a element set. The name of the character vector
#'     is the name of the element set.
#' @return For \code{tbl_elementset}, an S3 \code{elementset} object in a tibble
#'     representation.
#' @importFrom methods is
#' @importFrom tibble tibble
#' @importFrom dplyr distinct '%>%' select mutate group_by ungroup
#'     summarise arrange
#' @keywords internal
#' @examples
#' es <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)
.tbl_elementset <- function(...) {
    args <- list(...)
    if (length(args) == 1L && is(args[[1]], "list"))
        args <- args[[1]]

    stopifnot(
        all(vapply(args, is, logical(1), "character")),
        length(args) == 0 || !is.null(names(args)),
        all(nzchar(names(args)))
    )

    tbl <- tibble(
        element = as.character(unlist(args, use.names=FALSE)),
        set = as.character(rep(names(args), lengths(args)))
    )

    tbl <- tbl %>% distinct(.data$element, .data$set)

    subclass_tbl_elementset_base(tbl, "tbl_elementset")
}

#' @rdname tblelementset
#' @name tblelementset
#' @param x A tibble that may or may not be an elementset
#' @return For \code{.is_tbl_elementset} (internal), a logical
#'     indicating if the tibble is an elementset.
#' @keywords internal
#' @examples
#' 
#' BiocSet:::.is_tbl_elementset(es)
.is_tbl_elementset <- function(x) {
    all(c("element", "set") %in% names(x)) &&
        is.character(x$element) && is.character(x$set)
}

format.trunc_mat_tbl_elementset <- function(x, ...) {
    class <- sub("trunc_mat_", "", class(x)[1])
    names(x$summary) <- paste("A", class)
    NextMethod()
}

select.tbl_elementset <- function(.data, ...)
{
    tbl <- NextMethod("select", .data, .data$element, .data$set, ...)
    class(tbl) <- class(.data)
    tbl
}

context("tblelementset-class")

test_that("'.tbl_elementset()' works", {
    es <- .tbl_elementset(a = letters, B = LETTERS)
    expect_s3_class(es, "tbl_elementset")
    expect_identical(dim(es), c(52L,2L))
    expect_true(.is_tbl_elementset(es))

    es <- .tbl_elementset()
    expect_s3_class(es, "tbl_elementset")
    expect_identical(dim(es), c(0L,2L))
    expect_true(.is_tbl_elementset(es))

    es <- .tbl_elementset(a = character(), b = LETTERS)
    expect_s3_class(es, "tbl_elementset")
    expect_identical(dim(es), c(26L,2L))
    expect_true(.is_tbl_elementset(es))
    expect_identical(es$set, rep("b", 26))

    expect_error(.tbl_elementset(a = 1:5, B = LETTERS))
    expect_error(.tbl_elementset(a = 1:5))
    expect_error(.tbl_elementset(letters))
    expect_error(.tbl_elementset(a = letters, LETTERS))
})

test_that("'.tbl_elementset()' list() and named character vectors work", {
    es <- .tbl_elementset(list())
    expect_identical(es, .tbl_elementset())

    es <- .tbl_elementset(list(a = letters))
    expect_identical(es, .tbl_elementset(a = letters))

    es <- .tbl_elementset(list(a = letters, b = LETTERS))
    expect_identical(es, .tbl_elementset(a = letters, b = LETTERS))
})

test_that("'.is_tbl_elementset()' works", {
    expect_true(.is_tbl_elementset(tibble(element = character(),
                                         set = character())))
    expect_true(.is_tbl_elementset(tibble(element = character(),
                                         set = character(),
                                         x = integer())))
    expect_false(.is_tbl_elementset(tibble(element = integer(),
                                          set = factor())))
    expect_false(.is_tbl_elementset(tibble(element = character(),
                                          set = integer())))
    expect_false(.is_tbl_elementset(tibble()))
    expect_false(.is_tbl_elementset(tibble(element = character())))
})

context("tblelement-class")

test_that("'.tbl_element()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)

    es1 <- .tbl_element(es)
    expect_s3_class(es1, "tbl_element")
    expect_identical(class(es1), c("tbl_element",
                                   "tbl_elementset_base",
                                   "tbl_df",
                                   "tbl", "data.frame"))
    expect_length(class(es1), 5)
    expect_identical(es1$element, es$element)
    expect_identical(dim(es1), c(52L,1L))
    expect_false(.is_tbl_elementset(es1))
})

context("tblset-class")

test_that("'.tbl_set()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)

    es1 <- .tbl_set(es)
    expect_s3_class(es1, "tbl_set")
    expect_identical(class(es1), c("tbl_set",
                                   "tbl_elementset_base",
                                   "tbl_df",
                                   "tbl", "data.frame"))
    expect_length(class(es1), 5)
    expect_identical(dim(es1), c(2L,1L))
    expect_identical(levels(es1$set), levels(es$set))
    expect_false(.is_tbl_elementset(es1))
})

# This document is to show certain functionality of GSEABase and how the same
# functionality can be achieved (hopefully) with BiocSet.

# The first step is to acquire 
# http://software.broadinstitute.org/gsea/msigdb/download_file.jsp?filePath=/resources/msigdb/6.2/h.all.v6.2.symbols.gmt
# (this requires authentication)

# Load both of the packages...
library(GSEABase)
library(BiocSet)

# 1) ease of import of popular curated gene sets
# 1a) what are popular curated gene sets?

# I have kept the file mentioned above in my downloads, be sure to change the 
# path accordingly.
file <- "~/Downloads/h.all.v6.2.symbols.gmt"

# how to import and display with GSEABase
h1 = getGmt(file)
h1
head(names(h1))

# how to import and display with BiocSet
h2 = import(file)
h2
h2 %>% es_set() ## to show just the es_set() tibble
h2 %>% es_set() %>% pull(set) ## to show just the set column
h2 %>% es_set() %>% pull(set) %>% head() ## to show just the head of the sets

# Both packages are able to import the .gmt file and display the set names.

# 2) concept of "gene set collection" -- do we want to preserve it? clearly 
# msigdb implements the concept and that is a plus

# gene set collection with GSEABase
h1

# gene set collection with BiocSet
h2

# Though the information is not explicitly displayed in the BiocSet object, we 
# do show in the information within the tibbles. For example, the 'names:' in 
# h1 show a few examples and then (50 total). This information is shown in the
# es_set() tibble in h2 with the first 3 sets being printed and the dimensions 
# being 50x1. The same can be said for the 'unique identifiers:' in h1 being 
# shown in the es_element() tibble of h2.


# 2a) assuming we preseve collection concept, ease of measurement of set sizes

# set sizes with GSEABase
sapply(h1, function(x) length(geneIds(x)))

# set sizes with BiocSet
h2 %>% group_by(set) %>% dplyr::count()

# Both packages are able to give information about the set sizes.

# 2b) manage metadata about collections -- in GSEABase, the report on h1 above 
# is useful

# This information is not stored or displayed in the BiocSet object, but with 
# the `mutate_element()`, `mutate_set()`, or `mutate_elementset()` any info can
# be added to the tibbles.

# 3) manage metadata about sets

# metadata with GSEABase
details(h1[[1]])

# This most of this information is not stored or displayed in the BiocSet 
# object. I'm working on having the 'source' column in `es_elementset()` moved 
# to the `es_set()` tibble since this is mostly information regarding the sets.
# As I mentioned above with 2b), any of this information could be added by the 
# user if they needed it. 

# 4) support identifier conversions

# ID conversion with GSEABase
fl <- system.file("extdata", "Broad1.xml", package = "GSEABase")
bgs <- GeneSet(BroadCollection(), urls=fl)
bgs1 <- mapIdentifiers(bgs, AnnotationIdentifier("hgu95av2"))
bgs1

bgs2 <- mapIdentifiers(bgs, RefseqIdentifier("org.Hs.eg.db"))

# ID conversion with BiocSet (not ideal...)
library(hgu95av2.db)
library(org.Hs.eg.db)

xx <- as.list(hgu95av2SYMBOL)
genes <- geneIds(bgs)

bgs <- BiocSet("chr16q24" = unlist(xx[which(xx %in% genes)], use.names = FALSE))
bgs
bgs %>% es_map(hgu95av2.db, "SYMBOL", "PROBEID")

bgs %>% es_map(org.Hs.eg.db, "SYMBOL", "REFSEQ")

# One obvious issue here is that BiocSet does not support .xml files which is
# an area of improvement. I plan to take a look at what file types GSEABase
# supports and try to implement in BiocSet. I was able to do a bit of work 
# around using the 'hgu95av2.db' and 'org.Hs.eg.db' libraries to get a similar
# output to what Vince was able to produce with GSEABase. It's not a perfect
# match to Vince's output, but it gives a good idea of how BiocSet utilizes
# it's `es_map()` function to map between different IDs and databases.

# 5) relationships to ontologies -- a basic concern is relating cell types to 
# expression signatures, the latter are conceptually close to gene sets. This is
# a little table from the ctmarks app in ontoProc that starts to get at this...
# the basic idea is that some CL entries have exact-PRO relationships that we 
# can harvest to enumerate genes and thus generate gene sets. This seems like a
# good way to proceed to get institutionally endorsed assertions of cell type 
# signatures, but it goes slowly. Section 3.3 of the ontoProc vignette gets into
# ways of improvising extensions to CL to make use of the infrastructure in real
# time, anticipating proposals to CL maintainers as experimental evidence 
# solidifies.

# This would have to worked on a bit more. In our function `go_sets()` we do 
# allow the user to indicate with evidence type or ontology type they would like
# wen selecting the GO ids (the default is all evidence and ontoloty types). But
# this is as far as we got with working with ontologies.

.subclasses <- c("tbl_elementset_base",
                "tbl_element",
                "tbl_set",
                "tbl_elementset")

subclass_tbl_elementset_base <-
    function(x, subclass)
{
    class(x) <- c(subclass, "tbl_elementset_base", class(x))
    x
}

#' @importFrom dplyr filter
filter.tbl_elementset_base <-
    function(.data, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(.data)
    tbl
}

#' @importFrom dplyr select
select.tbl_elementset_base <-
    function(.data, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(.data)
    tbl
}

#' @importFrom dplyr mutate
mutate.tbl_elementset_base <-
    function(.data, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(.data)
    tbl
}

.tbl_nongroup_vars.tbl_elementset_base <-
    function(x)
{
    class = class(x)
    class(x) = setdiff(class, .subclasses)
    .tbl_nongroup_vars(x)
}

#' @importFrom dplyr summarise
summarise.tbl_elementset_base <-
    function(.data, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(.data)
    tbl
}

summarize.tbl_elementset_base <- summarise.tbl_elementset_base

#' @importFrom dplyr arrange
arrange.tbl_elementset_base <-
    function(.data, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(.data)
    tbl
}

#' @importFrom dplyr union
union.tbl_elementset_base <- function(x, y, ...)
{
    stopifnot(is(x) == is(y))
    tbl <- NextMethod()
    class(tbl) <- class(x)
    tbl
}

#' @importFrom dplyr intersect
intersect.tbl_elementset_base <- function(x, y, ...)
{
    stopifnot(is(x) == is(y))
    tbl <- NextMethod()
    class(tbl) <- class(x)
    tbl
}

#' @importFrom dplyr left_join
left_join.tbl_elementset_base <- function(x, y, by, copy, suffix, ...)
{
    tbl <- NextMethod()
    class(tbl) <- class(x)
    tbl
}

context("tbl_elementset_base-class")

test_that("'subclass_tbl_elementset_base()' works", {
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)

    es2 <- .tbl_element(es1)
    expect_s3_class(es2, "tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)

    es3 <- .tbl_set(es1)
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)
})

test_that("'filter.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- es %>% filter(element == "a")
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)


    es2 <- .tbl_element(es) %>% filter(element == "a" | element == "A")
    expect_s3_class(es2, "tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)

    es3 <- .tbl_set(es) %>% filter(set == "set1")
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)

    es_not <- es %>% filter(element == "1")
    expect_identical(dim(es_not), c(0L, 2L))
})

test_that("'select.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- es %>% select(element)
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)
    expect_error(es %>% select(Element))

    es2 <- .tbl_element(es) %>% select(element)
    expect_s3_class(es2, "tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)
    expect_error(.tbl_element(es) %>% select(set))

    es3 <- .tbl_set(es) %>% select(set)
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)
    expect_error(.tbl_set(es) %>% select(element))
})

test_that("'mutate.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- es %>% mutate(pval = rnorm(1:52))
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)
    expect_error(es %>% mutate(pval = rnorm(1:10)))

    es2 <- .tbl_element(es) %>% mutate(pval = rnorm(1:52))
    expect_s3_class(es2, "tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)
    expect_error(.tbl_element(es) %>% mutate(pval = rnorm(1:2)))

    es3 <- .tbl_set(es) %>% mutate(pval = rnorm(1:2))
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)
    expect_error(.tbl_set(es) %>% mutate(pval = rnorm(1:23)))
})

test_that("'.tbl_nongroup_vars.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)

    es1 <- es %>% .tbl_nongroup_vars()
    expect_is(es1, "character")
    expect_length(es1, 2)
    expect_error(es %>% .tbl_nongroup_vars(set))

    es2 <- .tbl_element(es) %>% .tbl_nongroup_vars()
    expect_is(es2, "character")
    expect_length(es2, 1)
    expect_error(es %>% .tbl_nongroup_vars(element))

    es3 <- .tbl_set(es) %>% .tbl_nongroup_vars()
    expect_is(es3, "character")
    expect_length(es3, 1)
    expect_error(es %>% .tbl_nongroup_vars(element))
})

test_that("'summarise.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- es %>% select(set) %>% summarise(n = n())
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)
    es1_2 <- es %>% select(set) %>% summarise(set)
    expect_identical(es1_2$set, es$set)

    es2 <- .tbl_element(es) %>% select(element) %>% summarise(n = n())
    expect_s3_class(es2,"tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)
    es2_2 <- es %>% select(element) %>% summarise(element)
    expect_identical(es2_2$element, es$element)

    es3 <- .tbl_set(es) %>% select(set) %>% summarise(n = n())
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)

    es4 <- es %>% select(set) %>% summarise()
    expect_identical(dim(es4), c(1L, 0L))
})

test_that("'arrange.tbl_elementset_base()' works", {
    es <- .tbl_elementset(set1 = letters, set2 = LETTERS)
    tbl <- c("tbl_elementset_base", "tbl_df", "tbl", "data.frame")

    es1 <- es %>% arrange(desc(element))
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", tbl))
    expect_length(class(es1), 5)
    expect_error(es %>% arrange(Element))

    es2 <- .tbl_element(es) %>% arrange(desc(element))
    expect_s3_class(es2, "tbl_element")
    expect_identical(class(es2), c("tbl_element", tbl))
    expect_length(class(es2), 5)
    expect_error(.tbl_element(es) %>% arrange(set))

    es3 <- .tbl_set(es) %>% arrange(desc(set))
    expect_s3_class(es3, "tbl_set")
    expect_identical(class(es3), c("tbl_set", tbl))
    expect_length(class(es3), 5)
    expect_error(.tbl_set(es) %>% arrange(element))
})

#' Functions applied to elements in a \code{BiocSet} object
#' @name element_funs
#' @rdname element_funs
#' @description All of the major methods applied to a \code{BiocSet} object can
#'     be explicitly applied to the element tibble. These functions bypass the
#'     need to use the \code{es_activate} function by indicating what function
#'     should be used on the element tibble.
#' @param .data A \code{BiocSet} object.
#' @param ... Additional arguments passed to the function.
#' @return A \code{BiocSet} object.
#' @export
#' @examples
#' es <- BiocSet(set1 = letters, set2 = LETTERS)
#' filter_element(es, element == "a")
filter_element <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "element") %>% filter(...)
    initialize(tbl, active = act)
}

#' @rdname element_funs
#'
#' @export
#'
#' @examples
#' 
#' es %>% select_element(element)
select_element <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "element") %>% select(...)
    initialize(tbl, active = act)
}

#' @rdname element_funs
#'
#' @export
#'
#' @examples
#' 
#' es %>% mutate_element(pval = rnorm(1:52))
mutate_element <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "element") %>% mutate(...)
    initialize(tbl, active = act)
}

#' @rdname element_funs
#'
#' @export
#'
#' @examples
#'
#' es %>% summarise_element(n = n())
summarise_element <- function(.data, ...) {
    act <- .active(.data)
    es_activate(.data, "element") %>% summarise(...)
}

#' @rdname element_funs
#'
#' @export
#'
#' @examples
#'
#' es %>% arrange_element(desc(element))
arrange_element <- function(.data, ...) {
    act <- .active(.data)
    tbl <- es_activate(.data, "element") %>% arrange(...)
    initialize(tbl, active = act)
}

#' @rdname element_funs
#'
#' @importFrom dplyr left_join
#'
#' @export
#'
#' @examples
#' 
#' tbl <- tibble(x = 1:5, y = letters[1:5])
#' es <- BiocSet(set1 = letters[c(1,3,5)], set2 = letters[c(2,4)])
#' left_join_element(es, tbl, by = c(element = "y"))
left_join_element <- function(.data, ...)
{
    tbl <- es_element(.data) %>% left_join(...)
    initialize(.data, element = tbl)
}

#' @rdname element_funs
#' @param how Multiple entries will become a list.
#' @return For \code{tibble_from_element}, a tibble.
#' @export
#' @examples
#' 
#' tibble_from_element(es)
tibble_from_element <-
    function(.data, how = unlist)
{
    element <- NULL
    tibble_from_elementset(.data) %>%
        group_by(element) %>%
        summarise_all(list) %>%
        mutate_if(.test, how)
}

#' @rdname element_funs
#' @return For \code{data.frame_from_element}, a data.frame.
#' @export
#' @examples
#' 
#' data.frame_from_element(es)
data.frame_from_element <-
    function(.data, how = unlist)
{
    tbl <- tibble_from_element(.data, how)
    data.frame(tbl, row.names = "element")
}

context("element_funs")

test_that("'tibble_from_element()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% tibble_from_element()
    expect_s3_class(es1, "tbl_df")
    expect_identical(dim(es1), c(52L, 2L))
    expect_true(.is_tbl_elementset(es1))
})

test_that("'data.frame_from_element()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% data.frame_from_element()
    expect_identical(class(es1), "data.frame")
    expect_identical(dim(es1), c(52L, 1L))
    expect_false(.is_tbl_elementset(es1))
})

#' Functions for mapping sets in the set tibble to different id types
#' @rdname mapping_set
#' @name mapping_set
#' @description Functions for creating \code{BiocSet} objects from GO sets and
#'     KEGG sets, and creating a new set mapping from a current \code{BiocSet} 
#'     object. \code{map_add_set} will add the mapping as a new column instead 
#'     of overwriting the current one used for the mapping.
#' @param org The AnnotationDbi object to identify keys/mappings from.
#' @param from A character to indicate which identifier to map from.
#' @param go A character to indicate the column name for the GO ids. 
#'     Default is "GO".
#' @param evidence A character to indicate the evidence codes for GO 
#'     associations with a gene of interest. Default is all possible evidence 
#'     codes.
#' @param ontology A character to indicate which Gene Ontology to use. 
#'     Default is BP, CC, and MF.
#' @importFrom AnnotationDbi keytypes keys
#' @return For \code{go_sets}, a \code{BiocSet} object with GO ids as the set 
#'     ids.
#' @export
#' @examples
#' library(org.Hs.eg.db)
#' go <- go_sets(org.Hs.eg.db, "ENSEMBL")
go_sets <- function(org, from, go = c("GO", "GOID"), evidence = NULL,
    ontology = NULL)
{
    stopifnot(from %in% keytypes(org))

    go <- match.arg(go)
    map <- AnnotationDbi::select(
        org, keys(org, from), c(from, go), from
    )
    map <- map[!is.na(map),]

    evidence_choices <- levels(as.factor(map$EVIDENCE))
    evid <- match.arg(evidence, evidence_choices, several.ok = TRUE)
    ontology_choices <- levels(as.factor(map$ONTOLOGY))
    onto <- match.arg(ontology, ontology_choices, several.ok = TRUE)

    if (!is.null(evidence) && !is.null(ontology)) {
        stopifnot(all(evid %in% evidence_choices),
            all(onto %in% ontology_choices))

        map <- filter(map, map$EVIDENCE %in% evid, map$ONTOLOGY %in% onto)
        evidence_split <- split(map$EVIDENCE, map[[go]])
        ontology_split <- split(map$ONTOLOGY, map[[go]])

        es <- do.call(BiocSet, split(map[[from]], map[[go]]))
        es %>% mutate_set(
            evidence = evidence_split,
            ontology = ontology_split
        )
    }
    else if (!is.null(evidence) && is.null(ontology)) {
        stopifnot(all(evidence %in% evidence_choices))

        map <- filter(map, map$EVIDENCE %in% evid)
        evidence_split <- split(map$EVIDENCE, map[[go]])
        evidence_unique <- lapply(evidence_split, unique)

        es <- do.call(BiocSet, split(map[[from]], map[[go]]))
        es %>% mutate_set(evidence = evidence_unique)
    }
    else if (is.null(evidence) && !is.null(ontology)) {
        stopifnot(all(ontology %in% ontology_choices))

        map <- filter(map, map$ONTOLOGY %in% onto)
        ontology_split <- split(map$ONTOLOGY, map[[go]])
        ontology_unique <- lapply(ontology_split, unique)

        es <- do.call(BiocSet, split(map[[from]], map[[go]]))
        es %>% mutate_set(ontology = ontology_unique)
    }
    else {
        do.call(BiocSet, split(map[[from]], map[[go]]))
    }
}

#' @rdname mapping_set
#' @name mapping_set
#' @param species Which species the pathways are from.
#' @importFrom KEGGREST keggList keggGet
#' @return For \code{kegg_sets}, a \code{BiocSet} object with Entrez IDs 
#'     reported as elements (default from KEGGREST) and KEGG pathways as sets.
#' @export
#' @examples
#'
#' kegg_sets("hsa")
kegg_sets <- function(species)
{
    stopifnot(species %in% keggList("organism")[,"organism"])

    name <- value <- NULL
    paths <- enframe(keggList("pathway", species))
    paths <- mutate(
        paths,
        name = gsub("path:", "", name),
        value = gsub("\\-.*", "", value)
        )

    grp <- cumsum(seq_along(paths$name) %% 10L == 1L)
    path_names <- split(paths$name, grp)
    elements <- lapply(path_names, function(x) {
        paths <- keggGet(x)
        lapply(paths, function(path) path$GENE[c(TRUE, FALSE)])
    })

    paths_len <- unlist(lapply(
        seq_along(elements), function(x) lengths(elements[[x]])
    ))
    elements <- unlist(elements, use.names = FALSE)
    elements <- split(elements, rep(seq_along(paths_len), paths_len))
    names(elements) <- paths$name[which(paths_len != 0)]

    do.call(BiocSet, elements)
}

#' @rdname mapping_set
#' @name mapping_set
#' @param .data The BiocSet object that contains the set tibble being mapped.
#' @param to A character to indicate which identifier to map to.
#' @return For \code{map_set}, a BiocSet object with the mapped set present in 
#'     the set tibble.
#' @export
#' @examples
#'
#' es <- BiocSet(set1 = letters, set2 = LETTERS)
#' es %>% map_set("set1", "foo") 
map_set <- function(.data, from, to) UseMethod("map_set")

#' @importFrom plyr mapvalues
#' @export
map_set.BiocSet <- function(.data, from, to)
{
    stopifnot(is.character(from), is.character(to), length(from) == length(to))

    set <- .set(.data)
    idx <- set$set %in% from
    set$set <- mapvalues(set$set, from, to)

    es <- .elementset(.data)
    idx <- es$set %in% from
    es$set <- mapvalues(es$set, from, to)

    initialize(.data, set = set, elementset = es)
}

#' @rdname mapping_set
#' @name mapping_set
#' @param add The id to add to the \code{BiocSet} object.
#' @return For \code{map_add_set}, a \code{BiocSet} object with a new column in
#'     the set tibble with the mapping of the new id type.
#' @export
#' @examples
#'
#' library(GO.db)
#' map <- map_add_set(go, GO.db, "GOID", "DEFINITION")
#' go %>% mutate_set(definition = map)
map_add_set <- function(.data, org, from, add)
{
    stopifnot(from %in% keytypes(org),
        add %in% keytypes(org))

    map <- mapIds(org,
        keys = as.character(es_set(.data)$set),
        column = add,
        keytype = from,
        multivals = "first"
    )
    unname(map)
}

context("mapping_set")

test_that("'go_sets()' works",
{
    library(org.Hs.eg.db)
    es <- go_sets(org.Hs.eg.db, "ENSEMBL")

    expect_s4_class(es, "BiocSet")
    expect_gte(dim(es_element(es))[1], 22432L)
    expect_identical(dim(es_element(es))[2], 1L)
    expect_gte(dim(es_set(es))[1], 18105L)
    expect_identical(dim(es_set(es))[2], 1L)
    expect_gte(dim(es_elementset(es))[1], 317035L)
    expect_identical(dim(es_elementset(es))[2], 2L)
    expect_true(.is_tbl_elementset(es_elementset(es)))

    es1 <- go_sets(org.Hs.eg.db, "SYMBOL", evidence = c("IDA", "IMP", "HDA"))

    expect_s4_class(es1, "BiocSet")
    expect_gte(dim(es_element(es1))[1], 13163L)
    expect_identical(dim(es_element(es1))[2], 1L)
    expect_gte(dim(es_set(es1))[1], 11069L)
    expect_identical(dim(es_set(es1))[2], 2L)
    expect_gte(dim(es_elementset(es1))[1], 87445L)
    expect_identical(dim(es_elementset(es1))[2], 2L)
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    
    es2 <- go_sets(org.Hs.eg.db, "SYMBOL", ontology = "MF")

    expect_s4_class(es2, "BiocSet")
    expect_gte(dim(es_element(es2))[1], 17696L)
    expect_identical(dim(es_element(es2))[2], 1L)
    expect_gte(dim(es_set(es2))[1], 4140L)
    expect_identical(dim(es_set(es2))[2], 2L)
    expect_gte(dim(es_elementset(es2))[1], 64664L)
    expect_identical(dim(es_elementset(es2))[2], 2L)
    expect_true(.is_tbl_elementset(es_elementset(es2)))

    es3 <- go_sets(org.Hs.eg.db, "ENSEMBL", evidence = c("IEP", "HEP"),
        ontology = c("CC", "MF"))

    expect_s4_class(es3, "BiocSet")
    expect_identical(dim(es_element(es3)), c(0L, 1L))
    expect_identical(dim(es_set(es3)), c(0L, 3L))
    expect_identical(dim(es_elementset(es3)), c(0L, 2L))
    expect_true(.is_tbl_elementset(es_elementset(es3)))

    expect_error(go_sets(org.Hs.eg.db))
    expect_error(go_sets(org.Hs.eg.db, ENSEMBL))
    expect_error(go_sets(org.Hs.eg.db, "IDS"))
    expect_error(go_sets(species, "ENSEMBL"))
    expect_error(go_sets(org.Hs.eg.db, "ENSEMBL", c("IDA", "IMP")))
    expect_error(go_sets(org.Hs.eg.db, "ENSEMBL", c("IDA", "IMP"), "CC"))
    expect_error(go_sets(org.Hs.eg.db, "ENSEMBL", "CC"))
    expect_error(go_sets(org.Hs.eg.db, "ENSEMBL", 1:2))
    expect_error(go_sets(org.Hs.eg.db, "SYMBOL", 1:2, 1:10))
    expect_error(go_sets(org.Hs.eg.db, "ENSEMBL", go))
    expect_error(go_sets())
})

test_that("'kegg_sets()' works",
{
    es <- kegg_sets("dme")

    expect_s4_class(es, "BiocSet")
    expect_identical(dim(es_element(es))[2], c(1L))
    expect_gte(dim(es_element(es))[1], c(3200L))
    expect_identical(dim(es_set(es))[2], c(1L))
    expect_gte(dim(es_set(es))[1], c(129L))
    expect_identical(dim(es_elementset(es))[2], c(2L))
    expect_gte(dim(es_elementset(es))[1], c(5340L))
    expect_true(.is_tbl_elementset(es_elementset(es)))

    expect_error(kegg_sets(hsa))
    expect_error(kegg_sets())
    expect_error(kegg_sets(1:2))
})

test_that("'map_set.BiocSet()' works", {
    es <- BiocSet(a = letters, B = LETTERS)

    es1 <- es %>% map_set("a", "A")
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_elementset(es1)), c(52L,2L))
    expect_identical(levels(es_elementset(es1)$set), levels(es_set(es1)$set))

    expect_error(es %>% map_set())
    expect_error(es %>% map_set("a"))
})

test_that("'map_add_set()' works",
{
    library(GO.db)
    go <- go_sets(org.Hs.eg.db, "ENSEMBL")
    map <- map_add_set(go, GO.db, "GOID", "DEFINITION")

    expect_gte(length(map), 18105L)
    expect_identical(class(map), "character")

    expect_error(map_add_set(GO.db, "GOID", "DEFINITION"))
})

#' Intersect on a single \code{BiocSet} object
#' @rdname intersect_single
#' @name intersect_single
#' @description This function performs an intersection within a single 
#'     \code{BiocSet} object.
#' @param x A \code{BiocSet} object.
#' @param ... Additional arguments passed to function.
#' @return A \code{BiocSet} object with a single set  'intersect' and 
#'     interesected elements from x.
#' @export
#' @examples
#' es1 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])
#' intersect_single(es1)
intersect_single <- function(x, ...)
{
    element <- NULL
    elements <- es_elementset(x)$element
    dup_list <- elements[duplicated(elements)]
    dup_elements <- x %>%
        filter_element(element %in% dup_list) %>%
        es_element() %>%
        pull(element)

    BiocSet(intersect = dup_elements)
}

context("intersect_single")

test_that("'intersect_single.BiocSet()' works",
{
    es1 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])

    tbl <- intersect_single(es1)
    expect_identical(dim(es_element(tbl)), c(7L, 1L))
    expect_identical(dim(es_set(tbl)), c(1L, 1L))
    expect_identical(dim(es_elementset(tbl)), c(7L, 2L)) 
})

.element <- function(x) x@element

.set <- function(x) x@set

.elementset <- function(x) x@elementset

.active <- function(x) x@active

`.active<-` <- function(x, value)
{
    value <- gsub('"', '', value)
    stopifnot(value %in% c("element", "set", "elementset"))
    x@active <- value
    x
}

#' @importFrom methods slot
.active_value <-
    function(x)
{
    slot(x, .active(x))
}

.test <-
    function(x)
{
    all(lengths(x) == 1L) &&
        length(unique(sapply(x, typeof))) == 1L
}

context("AllUtility")

test_that("'.element()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% es_element()
    expect_s3_class(es1, "tbl_element")
    expect_identical(class(es1), c("tbl_element", "tbl_elementset_base",
                                   "tbl_df", "tbl", "data.frame"))
    expect_length(class(es1), 5)
    expect_identical(dim(es1), c(52L,1L))
})

test_that("'.set()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% es_set()
    expect_s3_class(es1, "tbl_set")
    expect_identical(class(es1), c("tbl_set", "tbl_elementset_base",
                                  "tbl_df", "tbl", "data.frame"))
    expect_length(class(es1), 5)
    expect_identical(dim(es1), c(2L,1L))
})

test_that("'.elementset()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% es_elementset()
    expect_s3_class(es1, "tbl_elementset")
    expect_identical(class(es1), c("tbl_elementset", "tbl_elementset_base",
                                   "tbl_df", "tbl", "data.frame"))
    expect_length(class(es1), 5)
    expect_identical(dim(es1), c(52L,2L))
})

test_that("'.active()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% .active()
    expect_is(es1, "character")
    expect_length(es1, 1)
})

context("import-GMT")

test_that("'import.gmt()' works", {
    gmtFile <- system.file(package = "BiocSet", "extdata",
                           "hallmark.gene.symbol.gmt")
    es <- import.gmt(gmtFile)
    expect_true(.is_tbl_elementset(.elementset(es)))
    expect_identical(dim(.set(es)), c(50L, 2L))
    expect_identical(dim(.elementset(es)), c(7324L, 2L))
    expect_type(.set(es)$source, "character")
})

test_that("'import()' works", {
    gmtFile <- system.file(package = "BiocSet", "extdata",
                           "hallmark.gene.symbol.gmt")
    es <- import.gmt(gmtFile)
    fl <- tempfile(fileext = ".gmt")
    export(es, fl)
    expect_equal(es, import(fl))

    es <- BiocSet() %>% mutate_set(source = character(0))
    fl <- tempfile(fileext = ".gmt")
    export(es, fl)
    expect_identical(es, import(fl))

    es <- BiocSet(a = letters, b = LETTERS)
    fl <- tempfile(fileext = ".gmt")
    export(es, fl)
    expect_identical(es %>% mutate_set(source = NA_character_), import(fl))
})

#' Union on a single \code{BiocSet} object
#' @rdname union_single
#' @name union_single
#' @description This function performs a union within a single \code{BiocSet}
#'     object.
#' @param x A \code{BiocSet} object.
#' @param ... Additional arguments passed to function.
#' @importFrom dplyr pull
#' @return For \code{union_single}, a \code{BiocSet} object with a single set 
#'     \code{union} and unioned elements from x.
#' @export
#' @examples
#' es3 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])
#' union_single(es3)
union_single <- function(x, ...)
{
    element <- NULL
    unique_elements <- x %>% es_element() %>% pull(element)

    BiocSet(union = unique_elements)
}

#' @importFrom methods setOldClass
setOldClass(c("tbl_element", "tbl_set", "tbl_elementset"))

#' BiocSet class
#' @name BiocSet
#' @rdname BiocSet-class
#' @aliases BiocSet-class
#' @description character()
#' @slot element The element tibble from `tbl_elementset`
#' @slot set The set tibble from `tbl_elementset`
#' @slot elementset The elementset tibble created from user input
#' @slot active A character(1), indicates which tibble is active
#' @slot metadata A list() with arbitrary elements describing the set
#' @importClassesFrom S4Vectors Annotated
#' @importFrom S4Vectors metadata metadata<-
#' @export metadata metadata<-
#' @exportClass BiocSet
NULL

.BiocSet <- setClass(
    "BiocSet",
    contains = "Annotated",
    slots = c(
        element = "tbl_element",
        set = "tbl_set",
        elementset = "tbl_elementset",
        active = "character"
    )
)

## Constructor
#' @description The \code{BiocSet} constructor, the show method, the
#'     slot accessors, and creating a \code{BiocSet} object from an
#'     element set tibble rather than character vector(s).
#' @rdname BiocSet-class
#' @param ... Named character() vectors of element sets, or a named
#'     list of character() vectors. Each character vector is an
#'     element set. The names of the character vectors are the names of
#'     the sets.
#' @param metadata A list() with arbitrary content, describing the set.
#' @param active A character(1) to indicate which tibble is active. The
#'     default is "elementset".
#' @return An S4 \code{BiocSet} object shown as a tripple tibble,
#'     where each slot is a tibble.
#' @export
#' @examples
#' BiocSet(set1 = letters, set2 = LETTERS)
#' lst <- list(set1 = letters, set2 = LETTERS)
#' BiocSet(lst)
BiocSet <-
    function(..., metadata = list(), active = c("elementset", "element", "set"))
{
    active <- match.arg(active)
    elementset <- .tbl_elementset(...)
    element <- .tbl_element(elementset)
    set <- .tbl_set(elementset)

    .BiocSet(element = element,
             set = set,
             elementset = elementset,
             metadata = metadata,
             active = active)
}

#' @rdname BiocSet-class
#' @param object A \code{BiocSet} object.
#' @docType methods
setMethod(
    "show", "BiocSet",
    function(object)
    {
        active <- .active(object)
        cat("class: ", class(object), "\n", sep = "")
        cat("\nes_element()", if (active == "element")
                                " <active>", ":\n", sep = ""
            )
        print(.element(object), n = 3)
        cat("\nes_set()", if (active == "set")
                            " <active>", ":\n", sep = ""
            )
        print(.set(object), n = 3)
        cat("\nes_elementset()", if (active == "elementset")
                                    " <active>", ":\n", sep = ""
            )
        print(.elementset(object), n = 3)
    })

setGeneric(
    ".update",
    function(x, value) standardGeneric(".update"),
    signature = "value"
)

setMethod(
    ".update", "tbl_element",
    function(x, value)
{
    stopifnot(all(value$element %in% .element(x)$element))
    elementset <- filter(.elementset(x),
                        .elementset(x)$element %in% value$element)
    initialize(x, element = value, elementset = elementset)
})

setMethod(
    ".update", "tbl_set",
    function(x, value)
{
    stopifnot(all(value$set %in% .set(x)$set))
    elementset <- filter(.elementset(x), .elementset(x)$set %in% value$set)
    initialize(x, set = value, elementset = elementset)
})

setMethod(
    ".update", "tbl_elementset",
    function(x, value)
{
    stopifnot(
        all(value$element %in% .elementset(x)$element),
        all(value$set %in% .elementset(x)$set)
    )
    element <- filter(.element(x), .element(x)$element %in% value$element)
    set <- filter(.set(x), .set(x)$set %in% value$set)
    initialize(x, element = element, set = set, elementset = value)
})

update_es_element <- function(es, value)
    .update(es, value)

update_es_set <- function(es, value)
    .update(es, value)

update_es_elementset <- function(es, value)
    .update(es, value)

#' @rdname BiocSet-class
#' @param x A \code{BiocSet} object. 
#' @exportMethod es_element
setGeneric("es_element", function(x) standardGeneric("es_element"))

#' @rdname BiocSet-class
setMethod("es_element", "BiocSet", .element)

#' @rdname BiocSet-class
#' @exportMethod es_set
setGeneric("es_set", function(x) standardGeneric("es_set"))

#' @rdname BiocSet-class
setMethod("es_set", "BiocSet", .set)

#' @rdname BiocSet-class
#' @exportMethod es_elementset
setGeneric("es_elementset", function(x) standardGeneric("es_elementset"))

#' @rdname BiocSet-class
setMethod("es_elementset", "BiocSet", .elementset)

`es_element<-` <- update_es_element

`es_set<-` <- update_es_set

`es_elementset<-` <- update_es_elementset

#' @rdname BiocSet-class
#' @param elementset A tibble with element set information.
#' @param element A tibble with element information.
#' @param set A tibble with set information.
#' @export
#' @examples
#'
#' set.seed(123)
#' element <-
#'    tibble(
#'        element = letters[1:10],
#'        v1 = sample(10),
#'        v2 = sample(10)
#'    )
#' set <-
#'    tibble(
#'        set = LETTERS[1:2],
#'        v1 = sample(2),
#'        v2 = sample(2)
#'    )
#' elementset <-
#'    tibble(
#'        element = letters[1:10],
#'        set = sample(LETTERS[1:2], 10, TRUE)
#'    )
#' BiocSet_from_elementset(elementset, element, set)
BiocSet_from_elementset <- function(elementset, element, set, metadata)
{
    if (missing(elementset))
        elementset <- tibble(element = character(), set = character())
    if (missing(element))
        element <- tibble(element = character())
    if (missing(set))
        set <- tibble(set = character())
    if (missing(metadata))
        metadata <- list()
    stopifnot(
        "element" %in% names(elementset),
        is.character(elementset$element),
        "set" %in% names(elementset),
        "element" %in% names(element),
        is.character(element$element),
        "set" %in% names(set)
    )

    es <- do.call(
        BiocSet,
        c(
            split(elementset$element, elementset$set),
            list(metadata = metadata)
        )
    )
    es <- left_join_element(es, element, by = "element")
    es <- left_join_set(es, set, by = "set")
    es <- left_join_elementset(es, elementset, by = c("element", "set"))

    if (nrow(element) > nrow(es_element(es)))
        message("more elements in 'element' than in 'elementset'")
    if (nrow(set) > nrow(es_set(es)))
        message("more elements in 'set' than in 'elementset'")

    es
}

context("BiocSet-class")

test_that("'BiocSet()' works",
{
    es <- BiocSet(a = letters, b = LETTERS)
    expect_s4_class(es, "BiocSet")
    expect_identical(dim(es_element(es)), c(52L,1L))
    expect_identical(dim(es_set(es)), c(2L,1L))
    expect_identical(dim(es_elementset(es)), c(52L,2L))
    expect_true(.is_tbl_elementset(es_elementset(es)))
    expect_identical(.active(es), "elementset")
    expect_length(.active(es), 1L)

    es <- BiocSet()
    expect_s4_class(es, "BiocSet")
    expect_identical(dim(es_elementset(es)), c(0L,2L))
    expect_true(.is_tbl_elementset(es_elementset(es)))

    es <- BiocSet(set1 = character(), set2 = LETTERS)
    expect_s4_class(es, "BiocSet")
    expect_identical(dim(es_elementset(es)), c(26L,2L))
    expect_true(.is_tbl_elementset(es_elementset(es)))
    expect_identical(es_elementset(es)$set, rep("set2", 26))
    
    expect_error(BiocSet(set1 = 1:10, set2 = LETTERS))
    expect_error(BiocSet(set1 = 1:10))
    expect_error(BiocSet(LETTERS))
    expect_error(BiocSet(set1 = letters, LETTERS))
    expect_error(BiocSet(set1 = letters, set2 = 1:10))
    expect_error(BiocSet(1:10))
})

test_that("'BiocSet_from_elementset()' works", {
    expect_identical(BiocSet_from_elementset(), BiocSet())

    set.seed(123)
    element <- tibble(element = letters[1:10], v1 = sample(10), v2 = sample(10))
    set <- tibble(set = LETTERS[1:2], v1 = sample(2), v2 = sample(2))
    elementset <- 
        tibble(
            element = letters[1:10], 
            set = sample(LETTERS[1:2], 10, TRUE)
        )

    es <- BiocSet_from_elementset(elementset, element, set)
    expect_true(.is_tbl_elementset(es_elementset(es)))
    expect_identical(dim(es_element(es)), c(10L, 3L))
    expect_identical(dim(es_set(es)), c(2L, 3L))
    expect_identical(dim(es_elementset(es)), c(10L, 2L))

    es1 <- BiocSet_from_elementset(elementset)
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_element(es1)), c(10L, 1L))
    expect_identical(dim(es_set(es1)), c(2L, 1L))
    expect_identical(dim(es_elementset(es1)), c(10L, 2L))

    es2 <- BiocSet_from_elementset(elementset, element)
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_identical(dim(es_element(es2)), c(10L, 3L))
    expect_identical(dim(es_set(es2)), c(2L, 1L))
    expect_identical(dim(es_elementset(es2)), c(10L, 2L))

    es3 <- BiocSet_from_elementset(elementset, set = set)
    expect_true(.is_tbl_elementset(es_elementset(es3)))
    expect_identical(dim(es_element(es3)), c(10L, 1L))
    expect_identical(dim(es_set(es3)), c(2L, 3L))
    expect_identical(dim(es_elementset(es3)), c(10L, 2L))

    expect_error(BiocSet_from_elementset(elementset = 1:5))
})

.OBOFile = setClass("OBOFile", contains = "BiocFile")

OBOFile = function(resource, ...)
    .OBOFile(resource = resource)

## Need to account for diff format versions, order and tags can change
## These tags are for format version 1.2
.OBO_TAGS <- list(
    "1.2" = list(
        TERMS = c(
            "id", "is_anonymous", "name", "namespace", "alt_id", "def",
            "comment", "subset", "synonym", "xref", "is_a", "intersection_of",
            "union_of", "disjoint_from", "relationship", "is_obsolete",
            "replaced_by", "consider", "created_by", "creation_date"
        )
    )
)

.obo_term_tags <- function(version) {
    if (!version %in% names(.OBO_TAGS)) {
        warning("unknown OBO version '", version, "'; using version 1.2")
        version <- "1.2"
    }
    .OBO_TAGS[[version]]$TERMS
}

#' @importFrom ontologyIndex get_ontology
#' @importFrom tibble as_tibble tibble
#' @importFrom dplyr rename anti_join right_join
.import_obo <- function(path, extract_tags = "minimal") {
    stopifnot(
        extract_tags %in% c("minimal", "everything")
    )
    
    obo <- get_ontology(path, extract_tags = "everything")

    key <- NULL
    ## header
    version0 <- attr(obo, "version")
    pattern <- "^([^:]+):(.*)"
    obo_header <- tibble(
        key = sub(pattern, "\\1", version0),
        value = trimws(sub(pattern, "\\2", version0))
    )
    version <-
        obo_header %>%
        filter(key == "format-version") %>%
        pull("value")
    metadata <- list(obo_header = obo_header)

    ## parent / child ancestor information
    structure <- c("id", "parents", "children", "ancestors", "namespace")
    isa <-
        obo[intersect(names(obo), structure)] %>%
        lapply(function(x) unname(as.vector(x))) %>%
        as_tibble() %>%
        rename(element = "id")

    if ("namespace" %in% names(isa))
        isa <- isa %>% filter(.data$namespace != "external")

    ## elements
    minimal <- c("element", "name", "obsolete")
    tags <- c(minimal, .obo_term_tags(version))
    elements <-
        obo[intersect(names(obo), tags)] %>%
        lapply(function(x) unname(as.vector(x))) %>%
        as_tibble() %>%
        rename(element = "id")

    if ("namespace" %in% names(elements))
        elements <- elements %>% filter(.data$namespace != "external")

    if (extract_tags == "minimal")
        elements <- elements %>% select(minimal)

    ## sets
    sets <-
        elements %>%
        select(set = "element", "name")

    ## elementsets
    child <-
        isa %>%
        select(element = "element", set = "parents") %>%
        unnest("set") %>%
        mutate(is_a = TRUE)
    offspring <-
        isa %>%
        select(set = "ancestors", element = "element") %>%
        unnest("set")
    elementsets <-
        right_join(child, offspring, by = c("element", "set")) %>%
        mutate(is_a = !is.na(.data$is_a)) %>%
        select("element", "set", "is_a")

    OBOSet(elementsets, elements, sets, metadata)
}

#' @rdname import
#' @export
#' @examples
#'
#' oboFile <- system.file(package = "BiocSet", "extdata", "sample_go.obo")
#' tst_obo <- import(oboFile)
setMethod("import", c("OBOFile", "ANY", "ANY"),
    function(con, format, text, ...)
{
    .import_obo(resource(con), ...)
})

.element_elementset <- function(oboset) {
    is_a <- NULL
    inner_join(es_element(oboset),
        filter(es_elementset(oboset), is_a), 
        by = "element"
    )
}

.set_elementset <- function(oboset) {
    is_a <- NULL
    inner_join(es_set(oboset),
        filter(es_elementset(oboset), is_a),
        by = "set")
}

#' Functions to display relationships of an \code{OBOSet} object
#' @rdname obo_relations
#' @name obo_relations
#' @description These functions will display the relationships (children, 
#'     parents, or ancestors) for either the elements or the sets of an 
#'     \code{OBOSet} object.
#' @param oboset The \code{OBOSet} of interest.
#' @importFrom tidyr unnest
#' @importFrom dplyr inner_join
#' @return A 2 column tibble.
#' @export
#' @examples
#' oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")
#' obo <- import(oboFile)
#' oboset_element_children(obo) 
oboset_element_children <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    .element_elementset(oboset) %>%
        select("set", "element") %>%
        rename(element = "set", children = "element")
}

#' @rdname obo_relations
#' @export
#' @examples
#' 
#' oboset_element_parents(obo)
oboset_element_parents <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    .element_elementset(oboset) %>%
        select("element", "set") %>%
        rename(parents = "set")
}

#' @rdname obo_relations
#' @importFrom stats setNames
#' @export
#' @examples
#'
#' oboset_element_ancestors(obo)
oboset_element_ancestors <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    element <- NULL
    oboset %>%
        filter_element(element %in% es_elementset(oboset)$element) %>%
        .ancestors
}

#' @rdname obo_relations
#' @export
#' @examples
#'
#' oboset_set_children(obo)
oboset_set_children <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    .set_elementset(oboset) %>%
        select("set", "element") %>%
        rename(children = "element")
}

#' @rdname obo_relations
#' @export
#' @examples
#'
#' oboset_set_parents(obo)
oboset_set_parents <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    .set_elementset(oboset) %>%
        select("element", "set") %>%
        rename(set = "element", parents = "set")
}

#' @rdname obo_relations
#' @export
#' @examples
#'
#' oboset_set_ancestors(obo)
oboset_set_ancestors <- function(oboset) {
    stopifnot(is(oboset, "OBOSet"))
    set <- NULL
    oboset %>%
        filter_set(set %in% es_elementset(oboset)$set) %>%
        .ancestors %>%
        rename(set = "element")
}

.ancestors_merge <-
    function(ancestors, child, parent)
{
    ## ancestors of each child: the parent, and the parent's
    ## ancestors. A child is it's own ancestor
    ancestors0 <- Map(c, child, parent, ancestors[parent])
    ## a child may have more than one parent; unlist and re-split by child
    ancestors1 <- split(
        unlist(ancestors0, use.names = FALSE),
        rep(child, lengths(ancestors0))
    )
    
    ## combine newly identified ancestors with existing ancestors
    ## some newly identified ancestors may add to existing ancestors
    UFUN <- function(...) unique(c(...))
    duplicates <- intersect(names(ancestors), names(ancestors1))
    c(
        ancestors[setdiff(names(ancestors), duplicates)],
        ancestors1[setdiff(names(ancestors1), duplicates)],
        Map(UFUN, ancestors[duplicates], ancestors1[duplicates])
    )
}

.ancestors <-
    function(obo)
{
    stopifnot(is(obo, "OBOSet"))
    
    is_a <- NULL
    es <- es_elementset(obo) %>%
        filter(is_a) %>%
        select("element", "set") %>%
        distinct()
    
    child <- es %>% pull("element")
    parent <- es %>% pull("set")
    
    ancestor0 <- parent[!parent %in% child]
    root <- ancestor <- unique(ancestor0)
    ancestors <- setNames(as.list(root), root)
    
    total <- length(ancestor0)
    while (length(ancestor)) {
        ## index of parents who will beccome an ancestor this iteration
        idx <- parent %in% ancestor
        
        ## update list of ancestors
        ancestors <- .ancestors_merge(ancestors, child[idx], parent[idx])
        
        ## remove child-parent pairs of ancestors
        child <- child[!idx]
        parent <- parent[!idx]
        ancestor0 <- parent[!parent %in% child]
        ancestor <- unique(ancestor0)
    }
    
    ## all elements are there own ancestor
    child <-
        es_elementset(obo) %>%
        pull("element") %>%
        unique()
    ancestors <- .ancestors_merge(ancestors, child, child)
    
    ## tidy up
    ancestors <- lapply(ancestors, unique)
    tibble(
        element = rep(names(ancestors), lengths(ancestors)),
        ancestor = unlist(ancestors, use.names = FALSE)
    )
}

.oboset_is_a <- function(oboset) {
    . <- element <- is_a <- NULL
    isa <-
        es_elementset(oboset) %>%
        filter(.data$is_a) %>%
        select("element", "set")
    split(isa$set, isa$element) %>%
        tibble(element = names(.)) %>%
        select(element, is_a = ".") %>%
        mutate(is_a = unname(is_a))
}

.export_obo <- function(tbl, path) {
    stopifnot(
        .is_tbl_elementset(es_elementset(tbl)),
        `'path' exists` = !file.exists(path)
    )
   
    value <- NULL
    version <-
        metadata(tbl)[["obo_header"]] %>%
        filter(.data$key == "format-version") %>%
        pull(value)
    term_tags <- .obo_term_tags(version)

    elements <- es_element(tbl)

    if (!"is_a" %in% names(elements)) {
        isa <- .oboset_is_a(tbl)
        elements <- elements %>% left_join(isa, by = "element")
    }

    if ("obsolete" %in% names(elements) & is.logical(elements$obsolete)) {
        elements <-
            elements %>%
            rename(is_obsolete = "obsolete") %>%
            mutate(is_obsolete = as.integer(.data$is_obsolete))
    }

    if (!"id" %in% names(elements))
        elements <- elements %>% rename(id = "element")

    ## order elements by id
    elements <- elements %>% arrange(.data$id)

    ## keep only known columns
    elements <- elements[names(elements) %in% term_tags]

    ## term-key-value tibble
    kv <- tibble(
        term = rep(seq_len(nrow(elements)), ncol(elements)),
        key = rep(names(elements), each = nrow(elements)),
        value = unlist(elements, recursive = FALSE, use.names = FALSE)
    )
    ## 'unlist' value
    long <- tibble(
        term = rep(kv$term, lengths(kv$value)),
        key = rep(kv$key, lengths(kv$value)),
        value = unlist(kv$value)
    )

    tkv <-
        long %>%
        ## drop NAs
        filter(!is.na(.data$value)) %>%
        ## arrange by term then order in term_tags
        arrange(.data$term, factor(.data$key, levels = term_tags))

    records <- split(paste(tkv$key, tkv$value, sep=": "), tkv$term)

    ## output
    con <- file(path, "at")

    ## write header lines
    header <- metadata(tbl)$obo_header
    writeLines(paste0(header$key, ": ", header$value), con)

    ## write each record
    for (record in records)
        writeLines(c("", "[Term]", record), con)
    close(con)

    invisible(path)
}

#' @rdname import
#' @examples
#' fl <- system.file("extdata", "sample_go.obo", package = "BiocSet")
#' tbl <- import(fl)
#' new_fl <- tempfile(fileext = ".obo")
#' obo <- export(tbl, new_fl)
#' @export
setMethod(
    "export", c("BiocSet", "OBOFile", "ANY"),
    function(object, con, format, ...)
{
    .export_obo(object, resource(con))
})

context("import-OBO")

test_that("'.import_obo()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    ## minimal
    foo <- .import_obo(oboFile)
    expect_s4_class(foo, "OBOSet")
    expect_true(.is_tbl_elementset(.elementset(foo)))
    expect_identical(dim(.element(foo)), c(8L, 3L))
    expect_identical(dim(.set(foo)), c(8L, 2L))
    expect_identical(dim(.elementset(foo)), c(36L, 3L))

    element_nms <- c("element", "name", "obsolete")
    expect_true(all(colnames(.element(foo)) %in% element_nms))
    expect_type(.element(foo)$obsolete, "logical")

    set_nms <- c("set", "name")
    expect_true(all(colnames(.set(foo)) %in% set_nms))
    expect_type(.set(foo)$name, "character")

    elementset_nms <- c("element", "set", "is_a")
    expect_true(all(colnames(.elementset(foo)) %in% elementset_nms))
    expect_type(.elementset(foo)$is_a, "logical")
    
    ## everything
    foo_all <- .import_obo(oboFile, extract_tag = "everything")
    expect_s4_class(foo_all, "OBOSet")
    expect_true(.is_tbl_elementset(.elementset(foo_all)))
    expect_identical(dim(.element(foo_all)), c(8L, 14L))
    expect_identical(dim(.set(foo_all)), c(8L, 2L))
    expect_identical(dim(.elementset(foo_all)), c(36L, 3L))

    expect_type(.element(foo_all)$obsolete, "logical")
    expect_true(all(colnames(.set(foo_all)) %in% set_nms))
    expect_type(.set(foo_all)$name, "character")
    expect_true(all(colnames(.elementset(foo_all)) %in% elementset_nms))
    expect_type(.elementset(foo_all)$is_a, "logical")

    expect_error(.import_obo())
    expect_error(.import_obo(oboFile, all))
    expect_error(.import_obo(oboFile, extract_tag = all))
    expect_error(.import_obo(oboFile, extract_tag = "all"))
})

test_that("'import()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    ## minimal
    foo <- import(oboFile)
    expect_identical(foo, .import_obo(oboFile))

    ## everything
    foo_all <- import(oboFile, extract_tag = "everything")
    expect_identical(foo_all, .import_obo(oboFile, extract_tag = "everything"))

    expect_error(import())
    expect_error(import(oboFile, all))
    expect_error(import(oboFile, extract_tag = all))
    expect_error(import(oboFile, extract_tag = "all"))
})

test_that("'.element_elementset()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    foo <- import(oboFile)
    tst <- .element_elementset(foo)
    expect_true(is_tibble(tst))
    expect_identical(dim(tst), c(7L, 5L))
    names <- c("element", "name", "obsolete", "set", "is_a")
    expect_true(all(colnames(tst) %in% names))
    expect_true(all(tst$is_a))

    foo2 <- import(oboFile, extract_tag = "everything")
    tst2 <- .element_elementset(foo2)
    expect_identical(dim(tst2), c(7L, 16L))
})

test_that("'.set_elementset()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    foo <- import(oboFile)
    tst <- .set_elementset(foo)
    expect_true(is_tibble(tst))
    expect_identical(dim(tst), c(7L, 4L))
    names <- c("set", "name", "element", "is_a")
    expect_true(all(colnames(tst) %in% names))
    expect_true(all(tst$is_a))

    foo2 <- import(oboFile, extract_tag = "everything")
    tst2 <- .set_elementset(foo2)
    expect_identical(.set_elementset(foo2), tst)
})

test_that("'oboset_element_children()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_element_children(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(7L, 2L))
    names <- c("element", "children")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'oboset_element_parents()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_element_parents(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(7L, 2L))
    names <- c("element", "parents")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'obset_element_ancestors()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_element_ancestors(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(36L, 2L))
    names <- c("element", "ancestor")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'oboset_set_children()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_set_children(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(7L, 2L))
    names <- c("set", "children")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'oboset_set_parents()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_set_parents(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(7L, 2L))
    names <- c("set", "parents")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'oboset_set_ancestors()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    obo <- import(oboFile)
    tst <- oboset_set_ancestors(obo)
    expect_false(.is_tbl_elementset(tst))
    expect_identical(dim(tst), c(36L, 2L))
    names <- c("set", "ancestor")
    expect_true(all(colnames(tst) %in% names))
})

test_that("'.export_obo()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    ## minimal
    foo <- import(oboFile)
    fl <- tempfile(fileext=".obo")
    .export_obo(foo, fl)
    expect_equal(foo, import(fl))

    ## everything
    foo <- import(oboFile, extract_tag = "everything")
    fl <- tempfile(fileext=".obo")
    .export_obo(foo, fl)
    expect_equal(foo, import(fl, extract_tag = "everything"))

    foo2 <- foo %>% select_element(!obsolete)
    expect_error(.export_obo(foo2, fl))
})

test_that("'export()' works", {
    oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")

    ## minimal
    foo <- import(oboFile)
    fl <- tempfile(fileext=".obo")
    export(foo, fl)
    expect_equal(foo, import(fl))

    ## everything
    foo <- import(oboFile, extract_tag = "everything")
    fl <- tempfile(fileext=".obo")
    export(foo, fl)
    expect_equal(foo, import(fl, extract_tag = "everything"))

    foo2 <- foo %>% select_element(!obsolete)
    expect_error(export(foo2, fl))
})

#' GeneSetCollection
#' @rdname genesetcollection
#' @name genesetcollection
#' @description The following functions deal with converting a \code{BiocSet}
#'     object into a \code{GeneSetCollection} object, or vice versa.
#' @param biocset The \code{BiocSet} object that will become a
#'     \code{GeneSetCollection} object.
#' @return For `GeneSetCollection_from_BiocSet()`, a GeneSetCollection.
#' @export
#' @examples
#' biocset <- BiocSet(set1 = letters, set2 = LETTERS)
#' gsc <- GeneSetCollection_from_BiocSet(biocset)
#' gsc
GeneSetCollection_from_BiocSet <- function(biocset) {
    if(!requireNamespace("GSEABase", quietly = TRUE)) {
        stop("Please install GSEABase")
    }

    lst <- as.list(biocset)
    gsl <- Map(function(ids, name) {
        GSEABase::GeneSet(ids, setName = name)
    }, lst, names(lst))
    GSEABase::GeneSetCollection(gsl)
}

#' @rdname genesetcollection
#' @param gsc The \code{GeneSetCollection} that will become a \code{BiocSet}
#'     object.
#' @return For `BiocSet_from_GeneSetCollection()`, a BiocSet object.
#' @export
#' @examples
#'
#' BiocSet_from_GeneSetCollection(gsc)
BiocSet_from_GeneSetCollection <- function(gsc) {
    if(!requireNamespace("GSEABase", quietly = TRUE)) {
        stop("Please install GSEABase")
    }
    BiocSet(GSEABase::geneIds(gsc))
}

context("import-GeneSetCollection")

test_that("'GeneSetCollection_from_BiocSet()' works", {
    es <- BiocSet(set1 = letters[3:9], set2 = LETTERS[13:23])
    gsc <- GeneSetCollection_from_BiocSet(es)

    expect_s4_class(gsc, "GeneSetCollection")
    expect_identical(length(names(gsc)), 2L)
})

test_that("'BiocSet_from_GeneSetCollection()' works", {
    es <- BiocSet(set1 = letters[2:15], set2 = LETTERS[4:13])
    gsc <- GeneSetCollection_from_BiocSet(es)
    es2 <- BiocSet_from_GeneSetCollection(gsc)

    expect_s4_class(es2, "BiocSet")
    expect_identical(es2, es)
})

## This code shows how the sample_go.obo was created

## Start by downloading the LARGE go.obo file from GeneOntology
download.file("http://current.geneontology.org/ontology/go.obo", "obo_file.obo")

## Then use the import function to read in the obo file and create an OBOSet 
## object which is an extension of the BiocSet class
foo <- import("obo_file.obo", extract_tag = "everything")

## From here we created a list of ancestors to create a smaller obo file with 
## proper relationships
small_tst <- es_element(foo)[1,] %>%
    unnest("ancestors") %>%
    select("element", "ancestors") %>%
    unlist() %>%
    unique()

small_oboset <- foo %>% filter_elementset(element %in% small_tst)

## We demonstrate how to export to a temporary file but
## we exported the OBOset to the sample_go.obo file in inst/extdata of BiocSet
fl <- tempfile(fileext=".obo")
export(small_oboset, fl)

#' Element representation as an S3 class tibble
#' @rdname tblelement
#' @name tblelement
#' @param tbl_elementset An S3 \code{elementset} tibble.
#' @return An S3 \code{element} object in a tibble representation.
#' @importFrom tibble tibble
#' @importFrom dplyr distinct '%>%'
#' @keywords internal
#' @examples
#' tbl <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)
#' BiocSet:::.tbl_element(tbl)

.tbl_element <-
    function(tbl_elementset)
{
    stopifnot(
        .is_tbl_elementset(tbl_elementset)
    )
    
    tbl <- tibble(element = tbl_elementset$element) %>% distinct()
    subclass_tbl_elementset_base(tbl, "tbl_element")
}

.OBOSet_check_ancestry <- function(object) {
    errors <- character()

    if (!("children" %in% names(.set(object)))) {
        msg <- "'children' needs to be included in the obo file."
        errors <- c(errors, msg)
    }

    if (!("parents" %in% names(.set(object)))) {
        msg <- "'parents' needs to be included in the obo file."
        errors <- c(errors, msg)
    }

    if (!("ancestors" %in% names(.set(object)))) {
        msg <- "'ancestors' needs to be included in the obo file."
        errors <- c(errors, msg)
    }

    errors
}

#' @importFrom tibble is_tibble
.OBOSet_check_metadata <- function(object) {
    errors <- character()

    obo_header <- metadata(object)$obo_header
    if (!is_tibble(obo_header))
        errors <- c(errors, "'metadata$obo_header' must be a tibble")

    if (!all(c("key", "value") %in% names(obo_header))) {
        msg <- "'key', 'value' must be columns of 'metadata$obo_header"
        errors <- c(errors, msg)
    }

    errors
}

.OBOSet_validity <- function(object) {
    errors <- .OBOSet_check_ancestry(object)
    errors <- .OBOSet_check_metadata(object)

    if (length(errors)) errors else TRUE
}

#' @rdname OBOSet-class
#' @name OBOSet
#' @aliases OBOSet-class
#' @title OBOSet class
#' @description A class representing the 'OBO' file format as a BiocSet.
#' @exportClass OBOSet
NULL

.OBOSet <- setClass("OBOSet", contains = "BiocSet", validity = .OBOSet_validity)

#' @rdname OBOSet-class
#' @param elementset A tibble with element set information.
#' @param element A tibble with element information.
#' @param set A tibble with set information.
#' @param metadata A tibble with key-value pairs describing OBO file
#'     format header data
#' @return An S4 \code{OBOOSet} object. OBO sets conform to the 'obo'
#'     file format, with OBO 'Term' entries corresponding to
#'     elements. Parent / child relationships (e.g., 'is_a') are
#'     summarized as 'parents', 'ancestors', and 'children' character
#'     list columns of 'set'.
#' @examples
#' OBOSet()
#' oboFile <- system.file(package = "BiocSet", "extdata", "sample_go.obo")
#' import(oboFile)
#' @export
OBOSet <- function(elementset, element, set, metadata)
{
    if (missing(metadata)) {
        obo_header <- tibble(key = "format-version", value = "1.2")
        metadata <- list(obo_header = obo_header)
    }

    es <- BiocSet_from_elementset(elementset, element, set, metadata)
    .OBOSet(es)
}

.GMTFile = setClass("GMTFile", contains = "BiocFile")

GMTFile = function(resource, ...)
    .GMTFile(resource = resource)

#' @importFrom BiocIO import export resource
#' @importFrom methods new
#' @importFrom utils write.table
import.gmt <- function(path) {
    sets <- readLines(path)
    sets <- strsplit(sets, "\t")
    names <- vapply(sets, function(set) set[[1]], character(1))
    elements <- lapply(sets, function(set) set[-(1:2)])
    names(elements) <- names

    source <- vapply(sets, function(set) set[[2]], character(1))
    source[source=="NA" | !nzchar(source)] <- NA
    tbl <- do.call(BiocSet, elements)
    tbl <- tbl %>% mutate_set(source = source)
    tbl
}


#' Importing/exporting
#' @rdname import
#' @name import
#' @description Importing/exporting and formating of element sets as a 
#'     \code{BiocSet} object.
#' @param con For \code{import}, the file name or URL the element set is
#'     loaded from. For \code{export}, the file name or URL the element set is
#'     written to.
#' @param format For \code{import}, the format of the input. For 
#'     \code{export}, the format of the output.
#' @param text If con is missing this is a character vector directly providing 
#'     the element set that should be imported.
#' @param ... Parameters to pass to the format-specific method
#' @return For `import()`, a BiocSet object
#' @aliases import,GMTFile,ANY,ANY-method
#' @export
#' @examples
#' gmtFile <- system.file(package = "BiocSet", "extdata",
#'     "hallmark.gene.symbol.gmt")
#' tbl <- import(gmtFile)
setMethod(
    "import", c("GMTFile", "ANY", "ANY"),
    function(con, format, text, ...)
{
    import.gmt(resource(con))
})

#' @importFrom rlang .data
export.BiocSet <- function(tbl, path = tempfile(fileext = ".gmt")) {
    stopifnot(.is_tbl_elementset(es_elementset(tbl)))
    
    if(!"source" %in% names(es_set(tbl)))
        tbl <- mutate_set(tbl, source = rep(NA_character_, nrow(es_set(tbl))))

    es <- es_elementset(tbl)
    ## bug in dplyr
    if(nrow(es)==0L){
        sets <- tibble(set = character(0), element = character(0))
    } else {
        sets <- group_by(es, .data$set) %>%
            summarise(element = paste(.data$element, collapse = "\t"))
    }
    
    tbl <- left_join_set(tbl, sets, by = c(set = "set"))

    write.table(es_set(tbl), path, sep = "\t",
                col.names = FALSE, row.names = FALSE, quote = FALSE)
}

#' @rdname import
#'
#' @param object For `export()`, the object to be exported.
#'
#' @return For `export()`, a GMTFile object representing the location
#'     where the BiocSet object was written to
#'
#' @export
#'
#' @examples
#' 
#' tbl2 <- BiocSet(set1 = letters, set2 = LETTERS)
#' fl <- tempfile(fileext = ".gmt")
#' gmt <- export(tbl2, fl)
setMethod(
    "export", c("BiocSet", "GMTFile", "ANY"),
    function(object, con, format, ...)
{
    export.BiocSet(object, resource(con))
    con
})

#' BiocSet methods
#' @name BiocSet-methods
#' @rdname BiocSet-methods
#' @description \code{es_activate}: which of the three tibbles in the
#'     \code{BiocSet} object should be activated and have the chosen
#'     functionality applied to it.
#' @param .data The \code{BiocSet} object.
#' @param what Which of the three tibbles from \code{BiocSet} to activate.
#' @importFrom rlang quo_text enquo
#' @importFrom methods initialize
#' @return A \code{BiocSet} object.
#' @export
#' @examples
#' es <- BiocSet(set1 = letters, set2 = LETTERS)
#' es_activate(es, element)
es_activate <- function(.data, what)
{
    UseMethod("es_activate")
}

#' @export
#' @method es_activate BiocSet
es_activate.BiocSet <- function(.data, what)
{
    what <- quo_text(enquo(what))
    .active(.data) <- what
    .data
}

#' @rdname BiocSet-methods
#' @description \code{filter}: choose rows where conditions are true.
#' @param ... Additional arguments passed to function.
#' @export
#' @examples
#' 
#' es %>% es_activate(element) %>% filter(element == "a")
filter.BiocSet <- function(.data, ...)
{
    sub <- .active_value(.data)
    tbl <- filter(sub, ...)
    .update(.data, tbl)
}

#' @rdname BiocSet-methods
#' @description \code{select}: keep only the variables listed. 
#' @export
#' @examples
#' 
#' es %>% select(element)
select.BiocSet <- function(.data, ...)
{
    sub <- .active_value(.data)
    tbl <- select(sub, ...)
    .update(.data, tbl)
}

#' @rdname BiocSet-methods
#' @description \code{mutate}: add new variable and preserve the existing 
#'     variables.
#' @export
#' @examples
#'
#' es %>% es_activate(set) %>% mutate(pval = rnorm(1:2))
mutate.BiocSet <- function(.data, ...)
{
    stopifnot(!any(c("element", "set") %in% names(list(...))))

    sub <- .active_value(.data)
    tbl <- mutate(sub, ...)
    .update(.data, tbl)
}

#' @rdname BiocSet-methods
#' @description \code{summarise}: usually used with \code{group_by()}, output 
#'     will have one row for each group.
#' @export
#' @examples
#' 
#' es %>% es_activate(set) %>% summarise(n = n())
summarise.BiocSet <- function(.data, ...)
{
    sub <- .active_value(.data)
    summarise(sub, ...)
}

#' @rdname BiocSet-methods
#' @description \code{arrange}: order rows by an expression involving its 
#'     variables.
#' @export
#' @examples
#' 
#' es %>% es_activate(element) %>% arrange(desc(element))
arrange.BiocSet <- function(.data, ...)
{
    sub <- .active_value(.data)
    tbl <- arrange(sub, ...)
    .update(.data, tbl)
}

#' @importFrom dplyr tbl_nongroup_vars
.tbl_nongroup_vars <- tbl_nongroup_vars

#' @rdname BiocSet-methods
#' @description \code{.tbl_nongroup_vars}: returns only non-grouping variables.
#' @param x For \code{.tbl_nongroup_vars} (internal), a \code{BiocSet}
#'     object. For \code{union} and \code{intersect} the first
#'     \code{BiocSet} object to perform the operations on.
#' @keywords internal
#' @examples
#' 
#' es %>% mutate(pval = rnorm(1:52)) %>% es_elementset() %>%
#'     BiocSet:::.tbl_nongroup_vars()
.tbl_nongroup_vars.BiocSet <- function(x)
{
    active <- .active(x)
    sub <- slot(x, active)
    .tbl_nongroup_vars(sub)
}

#' @rdname BiocSet-methods
#' @description \code{group_by}: converts an existing tbl into a grouped tbl. 
#' @importFrom dplyr group_by
#' @param add logical, whether to add to the existing groups.
#' @export
#' @examples
#' 
#' es %>% group_by(element, set)
group_by.BiocSet <- function(.data, ..., add = FALSE)
{
    sub <- .active_value(.data)
    group_by(sub, ..., add = FALSE)
}

#' @rdname BiocSet-methods
#' @description \code{left_join}: returns all rows from \code{x}, and all 
#'     columns from \code{x} and \code{y}. If no rows in \code{x} match with 
#'     \code{y} there will be \code{NA}s in the new column. If there are 
#'     multiple matches then all combinations are returned.
#' @importFrom dplyr left_join
#' @param y For \code{left_join}, a tibble to join. For \code{union} and 
#'     \code{intersect} the second \code{BiocSet} object used.
#' @param by A character vector of variables to join by.
#' @param copy logical, allows you to join tables across srcs.
#' @param suffix Character vector of length 2, if there are non-joined duplicate
#'    variables in 'x' and 'y' these suffixes will be added to the output.
#' @export
#' @examples
#' 
#' es <- BiocSet(set1 = letters[1:5], set2 = LETTERS[1:5])
#' tbl <- tibble(x = 1:10, y = c(letters[1:5], LETTERS[1:5]))
#' es %>% left_join(tbl, by = c(element = "y"))
left_join.BiocSet <- function(x, y, by, copy, suffix, ...)
{
    sub <- .active_value(x)
        tbl <- left_join(sub, y = y, by = by, ...)
    .update(x, tbl)
}

.as.list.BiocSet <- function(from)
{
    element <- set <- NULL
    with(es_elementset(from), split(element, set))
}

#' @rdname BiocSet-methods
#' @description \code{as.list}: coerces argument into a list.
#' @export
#' @examples
#' 
#' library(org.Hs.eg.db)
#' es <- go_sets(org.Hs.eg.db, "ENSEMBL")
#' head(as.list(es))
as.list.BiocSet <- function(x, ...)
    .as.list.BiocSet(x)

#' as("BiocSet", "list")
#'
#' @name coerce
#' @aliases coerce,BiocSet,list-method
#' @importFrom methods setAs coerce
#' @exportMethod coerce
setAs("BiocSet", "list", .as.list.BiocSet)

#' @rdname BiocSet-methods
#' @description \code{union}: combines all rows from two \code{BiocSet} objects
#'     and removes duplicate records from the combined \code{BiocSet} object.
#' @importFrom dplyr union
#' @export
#' @examples
#' 
#' es1 <- BiocSet(set1 = letters[c(1:4)], set2 = LETTERS[c(1:4)])
#' es2 <- BiocSet(set1 = letters[c(3:8)], set2 = LETTERS[c(3:8)])
#'
#' dplyr::union(es1, es2)
union.BiocSet <- function(x, y, ...)
{
    element <- union(es_element(x), es_element(y), ...)
    set <- union(es_set(x), es_set(y), ...)
    elementset <- union(es_elementset(x), es_elementset(y), ...)
    initialize(x, element = element, set = set, elementset = elementset)
}

#' @rdname BiocSet-methods
#' @description \code{intersect}: combines all rows from two \code{BiocSet} 
#'     objects and returns rows that appear in both \code{BiocSet} objects.
#' @importFrom dplyr intersect
#' @export
#' @examples
#'
#' dplyr::intersect(es1, es2)
intersect.BiocSet <- function(x, y, ...)
{
    element <- intersect(es_element(x), es_element(y), ...)
    set <- intersect(es_set(x), es_set(y), ...)
    elementset <- intersect(es_elementset(x), es_elementset(y), ...)
    initialize(x, element = element, set = set, elementset = elementset)
}

context("BiocSet-methods")

test_that("'es_activate.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    expect_identical(.active(es), "elementset")

    es1 <- es %>% es_activate(element)
    expect_identical(.active(es1), "element")

    es2 <- es %>% es_activate(set)
    expect_identical(.active(es2), "set")

})

test_that("'filter.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)
    expect_equivalent(filter(es), es)

    es1 <- es %>% filter(set == "set1")
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_set(es1)), c(1L,1L))
    expect_identical(dim(es_elementset(es1)), c(26L,2L))
    expect_identical(es_elementset(es1)$element, letters)

    foo <- "k"
    es2 <- es %>% filter(element == foo)
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_identical(dim(es_element(es2)), c(1L,1L))
    expect_identical(es_elementset(es2)$element, foo)

    expect_error(es %>% filter(a))
})

test_that("'select.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% select(c(element, set))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_elementset(es1)), c(52L,2L))
    expect_identical(es1, es)

    es2 <- es %>% select(element)
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_identical(dim(es_elementset(es2)), c(52L,2L))
    expect_identical(es_element(es2), es_element(es))

    es3 <- es %>% select(set)
    expect_true(.is_tbl_elementset(es_elementset(es3)))
    expect_identical(dim(es_elementset(es3)), c(52L, 2L))
    expect_identical(es_set(es3), es_set(es))

    expect_error(es %>% select(a))
})

test_that("'mutate.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% mutate(pval = rnorm(1:52))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_elementset(es1)), c(52L,3L))

    expect_error(es %>% mutate(z = 1:2))
    expect_error(es %>% mutate(1:2))
    expect_error(es %>% mutate(a))
})

test_that("'summarise.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% summarise(n = n())
    expect_false(.is_tbl_elementset(es1))
    expect_identical(dim(es1), c(1L,1L))
    expect_equal(es1$n, 52L)

    es2 <- es %>% group_by(element) %>% summarise(n = n())
    expect_false(.is_tbl_elementset(es2))
    expect_identical(dim(es2), c(52L,2L))
    expect_equal(es2$n, c(rep(1L, 52)))

    es3 <- es %>% group_by(set) %>% summarise(n = n())
    expect_false(.is_tbl_elementset(es3))
    expect_identical(dim(es3), c(2L,2L))
    expect_equal(es3$n, c(26L, 26L))

    es4 <- es %>% group_by(set) %>% summarize(n = n())
    expect_false(.is_tbl_elementset(es4))
    expect_identical(dim(es4), c(2L, 2L))
    expect_equal(es4$n, c(26L, 26L))

    expect_error(es %>% summarise(a))
})

test_that("'arrange.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es %>% arrange()
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_equal(es1, es)

    es2 <- es %>% arrange(element)
    expect_true(.is_tbl_elementset(es_elementset(es2)))
    expect_equal(es_element(es2)$element, es_element(es)$element)
    expect_equal(es_elementset(es2)$element, sort(es_elementset(es)$element))

    es3 <- es %>% arrange(set)
    expect_true(.is_tbl_elementset(es_elementset(es3)))
    expect_equal(es3, es)
    expect_equal(es_elementset(es3)$set, sort(es_elementset(es)$set))

    es4 <- es %>% arrange(element == "k")
    expect_true(.is_tbl_elementset(es_elementset(es4)))
    expect_identical(dim(es_elementset(es4)), c(52L,2L))

    es5 <- es %>% arrange(set == "set1")
    expect_true(.is_tbl_elementset(es_elementset(es5)))
    expect_identical(dim(es_elementset(es5)), c(52L,2L))
})

test_that("'.tbl_nongroup_vars.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)

    es1 <- es_elementset(es) %>% .tbl_nongroup_vars()
    expect_identical(es1, c("element", "set"))
    expect_identical(length(es1), 2L)
    expect_identical(class(es1), "character")

    es2 <- es_element(es) %>% .tbl_nongroup_vars()
    expect_identical(es2, "element")
    expect_identical(length(es2), 1L)
    expect_identical(class(es2), "character")
})

test_that("'left_join.BiocSet()' works", {
    es <- BiocSet(set1 = letters, set2 = LETTERS)
    tbl <- tibble(x = rnorm(1:52), y = c(letters, LETTERS))

    es1 <- es %>% left_join(tbl, by = c(element = "y"))
    expect_true(.is_tbl_elementset(es_elementset(es1)))
    expect_identical(dim(es_elementset(es1)), c(52L, 3L))

    expect_error(es %>% left_join())
    expect_error(es %>% left_join(tbl))
    expect_error(es %>% left_join(tbl, by = 1:4))
    expect_error(es %>% left_join(tbl, by = "element"))
    expect_error(es %>% left_join(tbl, by = c("y" = element)))
})

test_that("'as.list.BiocSet()' works", {
    library(org.Hs.eg.db) 
    es <- go_sets(org.Hs.eg.db, "ENSEMBL")

    es1 <- as.list(es)
    expect_identical(class(es1), "list")
    expect_gte(length(es1), 18105L)
    expect_identical(names(es1), es_set(es)$set)
})

test_that("'union.BiocSet()' works",
{
    es1 <- BiocSet(set1 = letters[c(1:3)], set2 = LETTERS[c(1:3)])
    es2 <- BiocSet(set1 = letters[c(2:6)], set2 = LETTERS[c(2:6)])

    tbl <- union(es1, es2)
    expect_identical(dim(es_element(tbl)), c(12L, 1L))
    expect_identical(dim(es_set(tbl)), c(2L, 1L))
    expect_identical(dim(es_elementset(tbl)), c(12L, 2L))
})

test_that("'intersect.BiocSet()' works", 
{
    es1 <- BiocSet(set1 = letters[c(1:3)], set2 = LETTERS[c(1:3)])
    es2 <- BiocSet(set1 = letters[c(2:6)], set2 = LETTERS[c(2:6)])

    tbl <- intersect(es1, es2)
    expect_identical(dim(es_element(tbl)), c(4L, 1L))
    expect_identical(dim(es_set(tbl)), c(2L, 1L))
    expect_identical(dim(es_elementset(tbl)), c(4L, 2L))

})

---
title: "BiocSet: Representing Element Sets in the Tidyverse"
author: 
- name: Kayla Morrell
  affiliation: Roswell Park Comprehensive Cancer Center, Buffalo, NY
- name: Martin Morgan
  affiliation: Roswell Park Comprehensive Cancer Center, Buffalo, NY
date: "`r Sys.Date()`"
output:
    BiocStyle::html_document:
        toc: true
        toc_float: true
package: BiocSet
vignette: >
  %\VignetteIndexEntry{BiocSet: Representing Element Sets in the Tidyverse}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEndcoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

# Introduction
`BiocSet` is a package that represents element sets in a tibble format with the
 `BiocSet` class. Element sets are read in and converted into a tibble format. 
From here, typical `dplyr` operations can be performed on the element set. 
`BiocSet` also provides functionality for mapping different ID types and 
providing reference urls for elements and sets.

# Installation
Install the most recent version from Bioconductor:

```{r bioconductor, eval = FALSE}
if(!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("BiocSet")
```

The development version is also available for install from GitHub:

```{r github, eval = FALSE}
BiocManager::install("Kayla-Morrell/BiocSet")
```

Then load `BiocSet`:

```{r  load, message = FALSE}
library(BiocSet)
```
# BiocSet

## Input and Output

`BiocSet` can create a `BiocSet` object using two different input methods. The 
first is to input named character vectors of element sets. `BiocSet` returns 
three tibbles, `es_element` which contains the elements, `es_set` which contains
 the sets and `es_elementset` which contains elements and sets together. 

```{r constructor}
tbl <- BiocSet(set1 = letters, set2 = LETTERS)
tbl
```

The second method of creating a `BiocSet` object would be to read in a `.gmt` 
file. Using `import()`, a path to a downloaded `.gmt` file is read in and a 
`BiocSet` object is returned. The example below uses a hallmark element set 
downloaded from [GSEA][], which is also included with this package. This 
`BiocSet` includes a `source` column within the `es_elementset` tibble for 
reference as to where the element set came from.

[GSEA]: http://software.broadinstitute.org/esea/index.jsp

```{r gmt}
gmtFile <- system.file(package = "BiocSet",
                        "extdata",
                        "hallmark.gene.symbol.gmt")
tbl2 <- import(gmtFile)
tbl2
```

`export()` allows for a `BiocSet` object to be exported into a temporary file 
with the extention `.gmt`.

```{r export, tidy = TRUE}
fl <- tempfile(fileext = ".gmt")
gmt <- export(tbl2, fl)
gmt
```

We also support importing files to create an `OBOSet` object which extends 
the `BiocSet` class. The default only displays the most basic amount of columns 
needed to perform certain tasks, but the argument `extract_tag = everything` 
allows for additional columns to be imported as well. The following workflow 
will demonstrate the use of both export and import to create a smaller subset 
of the large obo file that is accessable from GeneOntology.

```{r obo, eval = FALSE}
download.file("http://current.geneontology.org/ontology/go.obo", "obo_file.obo")

foo <- import("obo_file.obo", extract_tag = "everything")

small_tst <- es_element(foo)[1,] %>%
    unnest("ancestors") %>%
    select("element", "ancestors") %>%
    unlist() %>%
    unique()

small_oboset <- foo %>% filter_elementset(element %in% small_tst)

fl <- tempfile(fileext = ".obo")
export(small_oboset, fl)
```

Then you can import this smaller obo file which we utilize here, in tests, and 
example code.
```{r small_obo}
oboFile <- system.file(package = "BiocSet",
                        "extdata",
                        "sample_go.obo")
obo <- import(oboFile)
obo
```

## Implemented functions

A feature available to `BiocSet` is the ability to activate different 
tibbles to perform certain functions on. When a `BiocSet` is created, the tibble
 `es_elementset` is automatically activated and all functions will be performed
 on this tibble. `BiocSet` adopts the use of many common `dplyr` functions such
 as `filter()`, `select()`, `mutate()`, `summarise()`, and `arrange()`. With 
each of the functions the user is able to pick a different tibble to activate
 and work on by using 'verb_tibble'. After the function is executed than the 
'active' tibble is returned back to the tibble that was active before the 
function call. Some examples are shown below of how these functions work.

```{r activate}
tbl <- BiocSet(set1 = letters, set2 = LETTERS)
tbl
tbl %>% filter_element(element == "a" | element == "A")
tbl %>% mutate_set(pval = rnorm(1:2))
tbl %>% arrange_elementset(desc(element))
```

## Set operations

`BiocSet` also allows for common set operations to be performed on the `BiocSet`
 object. `union()` and `intersection()` are the two set operations available in 
`BiocSet`. We demonstate how a user can find the union between two `BiocSet` 
objects or within a single `BiocSet` object. Intersection is used in the same 
way.

```{r set_ops}
# union of two BiocSet objects
es1 <- BiocSet(set1 = letters[c(1:3)], set2 = LETTERS[c(1:3)])
es2 <- BiocSet(set1 = letters[c(2:4)], set2 = LETTERS[c(2:4)])
union(es1, es2)

# union within a single BiocSet object
es3 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])
union_single(es3)
```

# Case study I

Next, we demonstrate the a couple uses of `BiocSet` with an experiment dataset 
`airway` from the package `airway`. This data is from an RNA-Seq experiment on 
airway smooth muscle (ASM) cell lines.

The first step is to load the library and the necessary data.

```{r airway, message = FALSE}
library(airway)
data("airway")
se <- airway
```

The function `go_sets()` discovers the keys from the org object and uses 
`AnnotationDbi::select` to create a mapping to GO ids. `go_sets()` also allows 
the user to indicate which evidence type or ontology type they would like when 
selecting the GO ids. The default is using all evidence types and all ontology 
types.  We represent these identifieres as a `BiocSet` object. Using the 
`go_sets` function we are able to map the Ensembl ids and GO ids from the genome
 wide annotation for Human data in the `org.Hs.eg.db` package. The Ensembl ids 
are treated as elements while the GO ids are treated as sets.

```{r go_sets, message = FALSE}
library(org.Hs.eg.db)
go <- go_sets(org.Hs.eg.db, "ENSEMBL")
go

# an example of subsetting by evidence type
go_sets(org.Hs.eg.db, "ENSEMBL", evidence = c("IPI", "TAS"))
```

Some users may not be interested in reporting the non-descriptive elements. We 
demonstrate subsetting the `airway` data to include non-zero assays and then 
filter out the non-descriptive elements.

```{r drop_assays}
se1 = se[rowSums(assay(se)) != 0,]
go %>% filter_element(element %in% rownames(se1))
```

It may also be of interest to users to know how many elements are in each set. 
Using the `count` function we are able to calculate the elements per set.

```{r count}
go %>% group_by(set) %>% dplyr::count()
```

It may also be helpful to remove sets that are empty. Since we have shown how 
to calculate the number of elements per set, we know that this data set does 
not contain any empty sets. We decide to demonstrate regardless for those users
 that may need this functionality.

```{r empty}
drop <- es_activate(go, elementset) %>% group_by(set) %>%
    dplyr::count() %>% filter(n == 0) %>% pull(set)
go %>% filter_set(!(set %in% drop))
```

To simplify mapping we created a couple `map` functions. `map_unique()` is used
 when there is known 1:1 mapping, This takes four arguements, a `BiocSet` 
object, an `AnnotationDbi` object, the id to map from, and the id to map to. 
`map_multiple()` needs a fifth argument to indicate how the function should 
treat an element when there is multiple mapping. Both functions utilize 
`mapIds` from `AnnotationDbi` and return a `BiocSet` object. In the example 
below we show how to use `map_unique` to map `go`'s ids from Ensembl to gene 
symbols.

```{r map_unique}
go %>% map_unique(org.Hs.eg.db, "ENSEMBL", "SYMBOL")
```

Another functionality of `BiocSet` is the ability to add information to the 
tibbles. Using the `GO.db` library we are able to map definitions to the GO 
ids. From there we can add the mapping to the tibble using `map_add()` and the 
mutate function.

```{r adding, message = FALSE}
library(GO.db)
map <- map_add_set(go, GO.db, "GOID", "DEFINITION")
go %>% mutate_set(definition = map)
```

The library `KEGGREST` is a client interface to the KEGG REST server. 
KEGG contains pathway maps that represent interaction, reaction and relation 
networks for various biological processes and diseases. `BiocSet` has a 
function that utilizes `KEGGREST` to develop a `BiocSet` object  that contains 
the elements for every pathway map in KEGG.

Due to limiations of the KEGGREST package, `kegg_sets` can take some time to 
run depending on the amount of pathways for the species of interest. Therefore 
we demonstrate using `BiocFileCache` to make the data available to the user.

```{r file_cache, message = FALSE}
library(BiocFileCache)
rname <- "kegg_hsa"
exists <- NROW(bfcquery(query=rname, field="rname")) != 0L
if (!exists)
{
    kegg <- kegg_sets("hsa")
    fl <- bfcnew(rname = rname, ext = ".gmt")
    export(kegg_sets("hsa"), fl)
}
kegg <- import(bfcrpath(rname=rname))
```

Within the `kegg_sets()` function we remove pathways that do not contain any 
elements. We then mutate the element tibble using the `map_add` function to 
contain both Ensembl and Entrez ids. 

```{r kegg_filter}
map <- map_add_element(kegg, org.Hs.eg.db, "ENTREZID", "ENSEMBL")
kegg <- kegg %>% mutate_element(ensembl = map)
```

Since we are working with ASM data we thought we would subset the `airway` data
 to contain only the elements in the asthma pathway. This filter is performed 
on the KEGG id, which for asthma is "hsa05310".

```{r subset}
asthma <- kegg %>% filter_set(set == "hsa05310")

se <- se[rownames(se) %in% es_element(asthma)$ensembl,]

se
```

The filtering can also be done for multiple pathways. 

```{r multiple}
pathways <- c("hsa05310", "hsa04110", "hsa05224", "hsa04970")
multipaths <- kegg %>% filter_set(set %in% pathways)

multipaths
```

# Case study II

This example will start out the same way that Case Study I started, by loading 
in the `airway` dataset, but we will also do some reformating of the data. The 
end goal is to be able to perform a Gene Set Enrichment Analysis on the data and
 return a BiocSet object of the gene sets.

```{r airway2, message = FALSE}
data("airway")
airway$dex <- relevel(airway$dex, "untrt")
```

Similar to other analyses we perform a differential expression analysis on the 
`airway` data using the library `DESeq2` and then store the results into a 
tibble.

```{r DE}
library(DESeq2)
library(tibble)
des <- DESeqDataSet(airway, design = ~ cell + dex)
des <- DESeq(des)
res <- results(des)

tbl <- res %>% 
    as.data.frame() %>%
    as_tibble(rownames = "ENSEMBL") 
```

Since we want to use `limma::goana()` to perform the GSEA we will need to have 
ENTREZ identifiers in the data, as well as filter out some `NA` information. 
Later on this will be considered our 'element' tibble.

```{r ENTREZ}
tbl <- tbl %>% 
    mutate(
        ENTREZID = mapIds(
            org.Hs.eg.db, ENSEMBL, "ENTREZID", "ENSEMBL"
        ) %>% unname()
    )

tbl <- tbl %>% filter(!is.na(padj), !is.na(ENTREZID))
tbl
```

Now that the data is ready for GSEA we can go ahead and use `goana()` and make 
the results into a tibble. This tibble will be considered our 'set' tibble.

```{r goana}
library(limma)
go_ids <- goana(tbl$ENTREZID[tbl$padj < 0.05], tbl$ENTREZID, "Hs") %>%
    as.data.frame() %>%
    as_tibble(rownames = "GOALL")
go_ids
```

The last thing we need to do is create a tibble that we will consider our 
'elementset' tibble. This tibble will be a mapping of all the elements and sets.

```{r final_tibble}
foo <- AnnotationDbi::select(
    org.Hs.eg.db,
    tbl$ENTREZID,
    "GOALL",
    "ENTREZID") %>% as_tibble()
foo <- foo %>% dplyr::select(-EVIDENCEALL) %>% distinct()
foo <- foo %>% filter(ONTOLOGYALL == "BP") %>% dplyr::select(-ONTOLOGYALL)
foo
```

The function `BiocSet_from_elementset()` allows for users to create a BiocSet 
object from tibbles. This function is helpful when there is metadata contained 
in the tibble that should be in the BiocSet object. For this function to work 
properly, the columns that are being joined on need to be named correctly. For 
instance, in order to use this function on the tibbles we created we need to 
change the column in the 'element' tibble to 'element', the column in the 'set' 
tibble to 'set' and the same will be for the 'elementset' tibble. We demonstrate
 this below and then create the BiocSet object with the simple function.

```{r BiocSet_from_elementset}
foo <- foo %>% dplyr::rename(element = ENTREZID, set = GOALL)
tbl <- tbl %>% dplyr::rename(element = ENTREZID)
go_ids <- go_ids %>% dplyr::rename(set = GOALL)
es <- BiocSet_from_elementset(foo, tbl, go_ids)
es
```

For those users that may need to put this metadata filled BiocSet object back 
into an object similar to GRanges or SummarizedExperiment, we have created 
functions that allow for the BiocSet object to be created into a tibble or 
data.frame.

```{r tibble_or_data.frame}
tibble_from_element(es)

head(data.frame_from_elementset(es))
```

# Information look up

A final feature to `BiocSet` is the ability to add reference information about 
all of the elements/sets. A user could utilize the function `url_ref()` to add 
information to the `BiocSet` object. If a user has a question about a specific 
id then they can follow the reference url to get more informtion. Below is an 
example of using `url_ref()` to add reference urls to the `go` data set we 
worked with above.

```{r url}
url_ref(go)
```

# Session info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/element_funs.R
\name{element_funs}
\alias{element_funs}
\alias{filter_element}
\alias{select_element}
\alias{mutate_element}
\alias{summarise_element}
\alias{arrange_element}
\alias{left_join_element}
\alias{tibble_from_element}
\alias{data.frame_from_element}
\title{Functions applied to elements in a \code{BiocSet} object}
\usage{
filter_element(.data, ...)

select_element(.data, ...)

mutate_element(.data, ...)

summarise_element(.data, ...)

arrange_element(.data, ...)

left_join_element(.data, ...)

tibble_from_element(.data, how = unlist)

data.frame_from_element(.data, how = unlist)
}
\arguments{
\item{.data}{A \code{BiocSet} object.}

\item{...}{Additional arguments passed to the function.}

\item{how}{Multiple entries will become a list.}
}
\value{
A \code{BiocSet} object.

For \code{tibble_from_element}, a tibble.

For \code{data.frame_from_element}, a data.frame.
}
\description{
All of the major methods applied to a \code{BiocSet} object can
    be explicitly applied to the element tibble. These functions bypass the
    need to use the \code{es_activate} function by indicating what function
    should be used on the element tibble.
}
\examples{
es <- BiocSet(set1 = letters, set2 = LETTERS)
filter_element(es, element == "a")

es \%>\% select_element(element)

es \%>\% mutate_element(pval = rnorm(1:52))

es \%>\% summarise_element(n = n())

es \%>\% arrange_element(desc(element))

tbl <- tibble(x = 1:5, y = letters[1:5])
es <- BiocSet(set1 = letters[c(1,3,5)], set2 = letters[c(2,4)])
left_join_element(es, tbl, by = c(element = "y"))

tibble_from_element(es)

data.frame_from_element(es)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OBOSet-class.R
\name{OBOSet}
\alias{OBOSet}
\alias{OBOSet-class}
\title{OBOSet class}
\usage{
OBOSet(elementset, element, set, metadata)
}
\arguments{
\item{elementset}{A tibble with element set information.}

\item{element}{A tibble with element information.}

\item{set}{A tibble with set information.}

\item{metadata}{A tibble with key-value pairs describing OBO file
format header data}
}
\value{
An S4 \code{OBOOSet} object. OBO sets conform to the 'obo'
    file format, with OBO 'Term' entries corresponding to
    elements. Parent / child relationships (e.g., 'is_a') are
    summarized as 'parents', 'ancestors', and 'children' character
    list columns of 'set'.
}
\description{
A class representing the 'OBO' file format as a BiocSet.
}
\examples{
OBOSet()
oboFile <- system.file(package = "BiocSet", "extdata", "sample_go.obo")
import(oboFile)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import-OBO.R
\name{obo_relations}
\alias{obo_relations}
\alias{oboset_element_children}
\alias{oboset_element_parents}
\alias{oboset_element_ancestors}
\alias{oboset_set_children}
\alias{oboset_set_parents}
\alias{oboset_set_ancestors}
\title{Functions to display relationships of an \code{OBOSet} object}
\usage{
oboset_element_children(oboset)

oboset_element_parents(oboset)

oboset_element_ancestors(oboset)

oboset_set_children(oboset)

oboset_set_parents(oboset)

oboset_set_ancestors(oboset)
}
\arguments{
\item{oboset}{The \code{OBOSet} of interest.}
}
\value{
A 2 column tibble.
}
\description{
These functions will display the relationships (children, 
    parents, or ancestors) for either the elements or the sets of an 
    \code{OBOSet} object.
}
\examples{
oboFile <- system.file("extdata", "sample_go.obo", package = "BiocSet")
obo <- import(oboFile)
oboset_element_children(obo) 

oboset_element_parents(obo)

oboset_element_ancestors(obo)

oboset_set_children(obo)

oboset_set_parents(obo)

oboset_set_ancestors(obo)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elementset_funs.R
\name{elementset_funs}
\alias{elementset_funs}
\alias{filter_elementset}
\alias{select_elementset}
\alias{mutate_elementset}
\alias{summarise_elementset}
\alias{arrange_elementset}
\alias{left_join_elementset}
\alias{tibble_from_elementset}
\alias{data.frame_from_elementset}
\title{Functions applied to elementsets in a \code{BiocSet} object}
\usage{
filter_elementset(.data, ...)

select_elementset(.data, ...)

mutate_elementset(.data, ...)

summarise_elementset(.data, ...)

arrange_elementset(.data, ...)

left_join_elementset(.data, ...)

tibble_from_elementset(.data)

data.frame_from_elementset(.data)
}
\arguments{
\item{.data}{A \code{BiocSet} object.}

\item{...}{Additional arguments passed to the function.}
}
\value{
A \code{BiocSet} object.

For \code{tibble_from_elementset}, a tibble.

For \code{data.frame_from_elementset}, a data.frame.
}
\description{
All of the major methods applied to a \code{BiocSet} object can
    be explicitly applied to the elementset tibble. These functions bypass 
    the need to use the \code{es_activate} function by indicating what 
    function should be used on the elementset tibble.
}
\examples{
es <- BiocSet(set1 = letters, set2 = LETTERS)
filter_elementset(es, element == "a" | element == "A")

es \%>\% select_elementset(element)

es \%>\% mutate_elementset(pval = rnorm(1:52))

es \%>\% summarise_elementset(n = n())

es \%>\% arrange_elementset(desc(element))

tbl <- tibble(x = 5:6, y = c("set1", "set2"))
es \%>\% left_join_elementset(tbl, by = c(set = "y"))

tibble_from_elementset(es)

data.frame_from_elementset(es)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_funs.R
\name{set_funs}
\alias{set_funs}
\alias{filter_set}
\alias{select_set}
\alias{mutate_set}
\alias{summarise_set}
\alias{arrange_set}
\alias{left_join_set}
\alias{tibble_from_set}
\alias{data.frame_from_set}
\title{Functions applied to sets in a \code{BiocSet} object}
\usage{
filter_set(.data, ...)

select_set(.data, ...)

mutate_set(.data, ...)

summarise_set(.data, ...)

arrange_set(.data, ...)

left_join_set(.data, ...)

tibble_from_set(.data, how = unlist)

data.frame_from_set(.data, how = unlist)
}
\arguments{
\item{.data}{A \code{BiocSet} object.}

\item{...}{Additional argument passed to the function.}

\item{how}{Multiple entries will become a list.}
}
\value{
A \code{BiocSet} object.

For \code{tibble_from_set}, a tibble.

For \code{data.frame_from_set}, a data.frame.
}
\description{
All of the major methods applied to a \code{BiocSet} object can
    be explicitly applied to the set tibble. These functions bypass the need
    to use the \code{es_activate} function by indicating what function should
    be used on the element tibble.
}
\examples{
es <- BiocSet(set1 = letters, set2 = LETTERS)
filter_set(es, set == "set1")

es \%>\% select_set(set)

es \%>\% mutate_set(pval = rnorm(1:2))

es \%>\% summarise_set(n = n())

es \%>\% arrange_set(desc(set))

tbl <- tibble(x = 10:11, y = c("set1", "set2"))
es <- BiocSet(set1 = letters[c(1,3,5)], set2 = letters[c(2,4)])
left_join_set(es, tbl, by = c(set = "y"))

tibble_from_set(es)

data.frame_from_set(es)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BiocSet-methods.R
\name{BiocSet-methods}
\alias{BiocSet-methods}
\alias{es_activate}
\alias{filter.BiocSet}
\alias{select.BiocSet}
\alias{mutate.BiocSet}
\alias{summarise.BiocSet}
\alias{arrange.BiocSet}
\alias{.tbl_nongroup_vars.BiocSet}
\alias{group_by.BiocSet}
\alias{left_join.BiocSet}
\alias{as.list.BiocSet}
\alias{union.BiocSet}
\alias{intersect.BiocSet}
\title{BiocSet methods}
\usage{
es_activate(.data, what)

\method{filter}{BiocSet}(.data, ...)

\method{select}{BiocSet}(.data, ...)

\method{mutate}{BiocSet}(.data, ...)

\method{summarise}{BiocSet}(.data, ...)

\method{arrange}{BiocSet}(.data, ...)

.tbl_nongroup_vars.BiocSet(x)

\method{group_by}{BiocSet}(.data, ..., add = FALSE)

\method{left_join}{BiocSet}(x, y, by, copy, suffix, ...)

\method{as.list}{BiocSet}(x, ...)

\method{union}{BiocSet}(x, y, ...)

\method{intersect}{BiocSet}(x, y, ...)
}
\arguments{
\item{.data}{The \code{BiocSet} object.}

\item{what}{Which of the three tibbles from \code{BiocSet} to activate.}

\item{...}{Additional arguments passed to function.}

\item{x}{For \code{.tbl_nongroup_vars} (internal), a \code{BiocSet}
object. For \code{union} and \code{intersect} the first
\code{BiocSet} object to perform the operations on.}

\item{add}{logical, whether to add to the existing groups.}

\item{y}{For \code{left_join}, a tibble to join. For \code{union} and 
\code{intersect} the second \code{BiocSet} object used.}

\item{by}{A character vector of variables to join by.}

\item{copy}{logical, allows you to join tables across srcs.}

\item{suffix}{Character vector of length 2, if there are non-joined duplicate
variables in 'x' and 'y' these suffixes will be added to the output.}
}
\value{
A \code{BiocSet} object.
}
\description{
\code{es_activate}: which of the three tibbles in the
    \code{BiocSet} object should be activated and have the chosen
    functionality applied to it.

\code{filter}: choose rows where conditions are true.

\code{select}: keep only the variables listed.

\code{mutate}: add new variable and preserve the existing 
    variables.

\code{summarise}: usually used with \code{group_by()}, output 
    will have one row for each group.

\code{arrange}: order rows by an expression involving its 
    variables.

\code{.tbl_nongroup_vars}: returns only non-grouping variables.

\code{group_by}: converts an existing tbl into a grouped tbl.

\code{left_join}: returns all rows from \code{x}, and all 
    columns from \code{x} and \code{y}. If no rows in \code{x} match with 
    \code{y} there will be \code{NA}s in the new column. If there are 
    multiple matches then all combinations are returned.

\code{as.list}: coerces argument into a list.

\code{union}: combines all rows from two \code{BiocSet} objects
    and removes duplicate records from the combined \code{BiocSet} object.

\code{intersect}: combines all rows from two \code{BiocSet} 
    objects and returns rows that appear in both \code{BiocSet} objects.
}
\examples{
es <- BiocSet(set1 = letters, set2 = LETTERS)
es_activate(es, element)

es \%>\% es_activate(element) \%>\% filter(element == "a")

es \%>\% select(element)

es \%>\% es_activate(set) \%>\% mutate(pval = rnorm(1:2))

es \%>\% es_activate(set) \%>\% summarise(n = n())

es \%>\% es_activate(element) \%>\% arrange(desc(element))

es \%>\% mutate(pval = rnorm(1:52)) \%>\% es_elementset() \%>\%
    BiocSet:::.tbl_nongroup_vars()

es \%>\% group_by(element, set)

es <- BiocSet(set1 = letters[1:5], set2 = LETTERS[1:5])
tbl <- tibble(x = 1:10, y = c(letters[1:5], LETTERS[1:5]))
es \%>\% left_join(tbl, by = c(element = "y"))

library(org.Hs.eg.db)
es <- go_sets(org.Hs.eg.db, "ENSEMBL")
head(as.list(es))

es1 <- BiocSet(set1 = letters[c(1:4)], set2 = LETTERS[c(1:4)])
es2 <- BiocSet(set1 = letters[c(3:8)], set2 = LETTERS[c(3:8)])

dplyr::union(es1, es2)

dplyr::intersect(es1, es2)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/url_ref.R
\name{url_ref}
\alias{url_ref}
\alias{url_ref_element}
\alias{url_ref_set}
\title{Functions to access reference urls for different identifiers}
\usage{
url_ref_element(es)

url_ref_set(es)

url_ref(es)
}
\arguments{
\item{es}{A \code{BiocSet} object that the reference urls should be added to.}
}
\value{
For \code{url_ref_element}, a \code{BiocSet} object with the url 
    column added to the element tibble.

For \code{url_ref_set}, a \code{BiocSet} object with the url column 
    added to the set tibble.

For \code{url_ref}, a \code{BiocSet} object with the url column 
    added to both the element and set tibbles.
}
\description{
Functions to access reference urls for different identifiers
}
\examples{
es <- BiocSet("GO:0000002" = c("TP53", "TNF"), "GO:0000003" = c("IL6"))
url_ref_element(es) 

url_ref_set(es)

url_ref(es)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import-GMT.R, R/import-OBO.R
\name{import}
\alias{import}
\alias{import,GMTFile,ANY,ANY-method}
\alias{export,BiocSet,GMTFile,ANY-method}
\alias{import,OBOFile,ANY,ANY-method}
\alias{export,BiocSet,OBOFile,ANY-method}
\title{Importing/exporting}
\usage{
\S4method{import}{GMTFile,ANY,ANY}(con, format, text, ...)

\S4method{export}{BiocSet,GMTFile,ANY}(object, con, format, ...)

\S4method{import}{OBOFile,ANY,ANY}(con, format, text, ...)

\S4method{export}{BiocSet,OBOFile,ANY}(object, con, format, ...)
}
\arguments{
\item{con}{For \code{import}, the file name or URL the element set is
loaded from. For \code{export}, the file name or URL the element set is
written to.}

\item{format}{For \code{import}, the format of the input. For 
\code{export}, the format of the output.}

\item{text}{If con is missing this is a character vector directly providing 
the element set that should be imported.}

\item{...}{Parameters to pass to the format-specific method}

\item{object}{For `export()`, the object to be exported.}
}
\value{
For `import()`, a BiocSet object

For `export()`, a GMTFile object representing the location
    where the BiocSet object was written to
}
\description{
Importing/exporting and formating of element sets as a 
    \code{BiocSet} object.
}
\examples{
gmtFile <- system.file(package = "BiocSet", "extdata",
    "hallmark.gene.symbol.gmt")
tbl <- import(gmtFile)

tbl2 <- BiocSet(set1 = letters, set2 = LETTERS)
fl <- tempfile(fileext = ".gmt")
gmt <- export(tbl2, fl)

oboFile <- system.file(package = "BiocSet", "extdata", "sample_go.obo")
tst_obo <- import(oboFile)
fl <- system.file("extdata", "sample_go.obo", package = "BiocSet")
tbl <- import(fl)
new_fl <- tempfile(fileext = ".obo")
obo <- export(tbl, new_fl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BiocSet-methods.R
\name{coerce}
\alias{coerce}
\alias{coerce,BiocSet,list-method}
\title{as("BiocSet", "list")}
\description{
as("BiocSet", "list")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intersect_single.R
\name{intersect_single}
\alias{intersect_single}
\title{Intersect on a single \code{BiocSet} object}
\usage{
intersect_single(x, ...)
}
\arguments{
\item{x}{A \code{BiocSet} object.}

\item{...}{Additional arguments passed to function.}
}
\value{
A \code{BiocSet} object with a single set  'intersect' and 
    interesected elements from x.
}
\description{
This function performs an intersection within a single 
    \code{BiocSet} object.
}
\examples{
es1 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])
intersect_single(es1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/union_single.R
\name{union_single}
\alias{union_single}
\title{Union on a single \code{BiocSet} object}
\usage{
union_single(x, ...)
}
\arguments{
\item{x}{A \code{BiocSet} object.}

\item{...}{Additional arguments passed to function.}
}
\value{
For \code{union_single}, a \code{BiocSet} object with a single set 
    \code{union} and unioned elements from x.
}
\description{
This function performs a union within a single \code{BiocSet}
    object.
}
\examples{
es3 <- BiocSet(set1 = letters[c(1:10)], set2 = letters[c(4:20)])
union_single(es3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping_set.R
\name{mapping_set}
\alias{mapping_set}
\alias{go_sets}
\alias{kegg_sets}
\alias{map_set}
\alias{map_add_set}
\title{Functions for mapping sets in the set tibble to different id types}
\usage{
go_sets(org, from, go = c("GO", "GOID"), evidence = NULL, ontology = NULL)

kegg_sets(species)

map_set(.data, from, to)

map_add_set(.data, org, from, add)
}
\arguments{
\item{org}{The AnnotationDbi object to identify keys/mappings from.}

\item{from}{A character to indicate which identifier to map from.}

\item{go}{A character to indicate the column name for the GO ids. 
Default is "GO".}

\item{evidence}{A character to indicate the evidence codes for GO 
associations with a gene of interest. Default is all possible evidence 
codes.}

\item{ontology}{A character to indicate which Gene Ontology to use. 
Default is BP, CC, and MF.}

\item{species}{Which species the pathways are from.}

\item{.data}{The BiocSet object that contains the set tibble being mapped.}

\item{to}{A character to indicate which identifier to map to.}

\item{add}{The id to add to the \code{BiocSet} object.}
}
\value{
For \code{go_sets}, a \code{BiocSet} object with GO ids as the set 
    ids.

For \code{kegg_sets}, a \code{BiocSet} object with Entrez IDs 
    reported as elements (default from KEGGREST) and KEGG pathways as sets.

For \code{map_set}, a BiocSet object with the mapped set present in 
    the set tibble.

For \code{map_add_set}, a \code{BiocSet} object with a new column in
    the set tibble with the mapping of the new id type.
}
\description{
Functions for creating \code{BiocSet} objects from GO sets and
    KEGG sets, and creating a new set mapping from a current \code{BiocSet} 
    object. \code{map_add_set} will add the mapping as a new column instead 
    of overwriting the current one used for the mapping.
}
\examples{
library(org.Hs.eg.db)
go <- go_sets(org.Hs.eg.db, "ENSEMBL")

kegg_sets("hsa")

es <- BiocSet(set1 = letters, set2 = LETTERS)
es \%>\% map_set("set1", "foo") 

library(GO.db)
map <- map_add_set(go, GO.db, "GOID", "DEFINITION")
go \%>\% mutate_set(definition = map)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tblelementset-class.R
\name{tblelementset}
\alias{tblelementset}
\alias{.tbl_elementset}
\alias{.is_tbl_elementset}
\title{Element set representation as an S3 class tibble}
\usage{
.tbl_elementset(...)

.is_tbl_elementset(x)
}
\arguments{
\item{...}{For \code{tbl_elementset}, named character vectors of gene sets. 
Each character vector is a element set. The name of the character vector
is the name of the element set.}

\item{x}{A tibble that may or may not be an elementset}
}
\value{
For \code{tbl_elementset}, an S3 \code{elementset} object in a tibble
    representation.

For \code{.is_tbl_elementset} (internal), a logical
    indicating if the tibble is an elementset.
}
\description{
Element set representation as an S3 class tibble
}
\examples{
es <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)

BiocSet:::.is_tbl_elementset(es)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import-GeneSetCollection.R
\name{genesetcollection}
\alias{genesetcollection}
\alias{GeneSetCollection_from_BiocSet}
\alias{BiocSet_from_GeneSetCollection}
\title{GeneSetCollection}
\usage{
GeneSetCollection_from_BiocSet(biocset)

BiocSet_from_GeneSetCollection(gsc)
}
\arguments{
\item{biocset}{The \code{BiocSet} object that will become a
\code{GeneSetCollection} object.}

\item{gsc}{The \code{GeneSetCollection} that will become a \code{BiocSet}
object.}
}
\value{
For `GeneSetCollection_from_BiocSet()`, a GeneSetCollection.

For `BiocSet_from_GeneSetCollection()`, a BiocSet object.
}
\description{
The following functions deal with converting a \code{BiocSet}
    object into a \code{GeneSetCollection} object, or vice versa.
}
\examples{
biocset <- BiocSet(set1 = letters, set2 = LETTERS)
gsc <- GeneSetCollection_from_BiocSet(biocset)
gsc

BiocSet_from_GeneSetCollection(gsc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping_element.R
\name{mapping_element}
\alias{mapping_element}
\alias{map_unique}
\alias{map_multiple}
\alias{map_add_element}
\title{Functions for mapping elements in the element tibble to different id types}
\usage{
map_unique(es, org, from, to)

map_multiple(
  es,
  org,
  from,
  to,
  multi = c("list", "filter", "asNA", "CharacterList")
)

map_add_element(es, org, from, add)
}
\arguments{
\item{es}{The BiocSet objec to map the elements on.}

\item{org}{The AnnotationDbi object to identify keys/mappings from.}

\item{from}{A character to indicate which identifier to map from.}

\item{to}{A character to indicate which identifier to map to.}

\item{multi}{How should multiple values be returned? 
Options include:     
\itemize{
\item{list: This will just return a list object to the end user.}
\item{filter: This will remove all elements that contain multiple 
matches and will therefore return a shorter vector than what came in 
whenever some of the keys match more than one value.}
\item{asNA: This will return an NA value whenever there are multiple 
matches.}
\item{CharacterList: This just returns a SimpleCharacterList object.}
\item{FUN: A function can be supplied to the 'multiVals' argument 
for custom behaviors.}
}}

\item{add}{The id to add to the \code{BiocSet} object.}
}
\value{
For \code{map_unique}, a \code{BiocSet} object with unique 
   elements.

For \code{map_multiple}, a \code{BiocSet} object with multiple 
    mappings for certain elements.

For \code{map_add_element}, a \code{BiocSet} object with a new column
    in the element tibble with the mapping of the new id type.
}
\description{
Functions for dealing with unique mapping and multiple mapping.
    \code{map_add_element} will add the mapping as a new column instead of 
    overwriting the current one used for the mapping.
}
\examples{
library(org.Hs.eg.db)
es <- BiocSet(set1 = c("C5", "GANC"), set2 = c("AFM", "CGB1", "ADAM32"))
map_unique(es, org.Hs.eg.db, "SYMBOL", "ENTREZID")

map_multiple(es, org.Hs.eg.db, "SYMBOL", "ENSEMBLTRANS", "asNA")

map <- map_add_element(es, org.Hs.eg.db, "SYMBOL", "ENTREZID")
es \%>\% mutate_element(entrez = map)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BiocSet-class.R
\docType{methods}
\name{BiocSet}
\alias{BiocSet}
\alias{BiocSet-class}
\alias{show,BiocSet-method}
\alias{es_element}
\alias{es_element,BiocSet-method}
\alias{es_set}
\alias{es_set,BiocSet-method}
\alias{es_elementset}
\alias{es_elementset,BiocSet-method}
\alias{BiocSet_from_elementset}
\title{BiocSet class}
\usage{
BiocSet(..., metadata = list(), active = c("elementset", "element", "set"))

\S4method{show}{BiocSet}(object)

es_element(x)

\S4method{es_element}{BiocSet}(x)

es_set(x)

\S4method{es_set}{BiocSet}(x)

es_elementset(x)

\S4method{es_elementset}{BiocSet}(x)

BiocSet_from_elementset(elementset, element, set, metadata)
}
\arguments{
\item{...}{Named character() vectors of element sets, or a named
list of character() vectors. Each character vector is an
element set. The names of the character vectors are the names of
the sets.}

\item{metadata}{A list() with arbitrary content, describing the set.}

\item{active}{A character(1) to indicate which tibble is active. The
default is "elementset".}

\item{object}{A \code{BiocSet} object.}

\item{x}{A \code{BiocSet} object.}

\item{elementset}{A tibble with element set information.}

\item{element}{A tibble with element information.}

\item{set}{A tibble with set information.}
}
\value{
An S4 \code{BiocSet} object shown as a tripple tibble,
    where each slot is a tibble.
}
\description{
character()

The \code{BiocSet} constructor, the show method, the
    slot accessors, and creating a \code{BiocSet} object from an
    element set tibble rather than character vector(s).
}
\section{Slots}{

\describe{
\item{\code{element}}{The element tibble from `tbl_elementset`}

\item{\code{set}}{The set tibble from `tbl_elementset`}

\item{\code{elementset}}{The elementset tibble created from user input}

\item{\code{active}}{A character(1), indicates which tibble is active}

\item{\code{metadata}}{A list() with arbitrary elements describing the set}
}}

\examples{
BiocSet(set1 = letters, set2 = LETTERS)
lst <- list(set1 = letters, set2 = LETTERS)
BiocSet(lst)

set.seed(123)
element <-
   tibble(
       element = letters[1:10],
       v1 = sample(10),
       v2 = sample(10)
   )
set <-
   tibble(
       set = LETTERS[1:2],
       v1 = sample(2),
       v2 = sample(2)
   )
elementset <-
   tibble(
       element = letters[1:10],
       set = sample(LETTERS[1:2], 10, TRUE)
   )
BiocSet_from_elementset(elementset, element, set)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tblset-class.R
\name{tblset}
\alias{tblset}
\alias{.tbl_set}
\title{Set representation as an S3 class tibble}
\usage{
.tbl_set(tbl_elementset)
}
\arguments{
\item{tbl_elementset}{An S3 \code{elementset} tibble.}
}
\value{
An S3 \code{set} object in a tibble representation.
}
\description{
Set representation as an S3 class tibble
}
\examples{
tbl <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)
BiocSet:::.tbl_set(tbl)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tblelement-class.R
\name{tblelement}
\alias{tblelement}
\alias{.tbl_element}
\title{Element representation as an S3 class tibble}
\usage{
.tbl_element(tbl_elementset)
}
\arguments{
\item{tbl_elementset}{An S3 \code{elementset} tibble.}
}
\value{
An S3 \code{element} object in a tibble representation.
}
\description{
Element representation as an S3 class tibble
}
\examples{
tbl <- BiocSet:::.tbl_elementset(set1 = letters, set2 = LETTERS)
BiocSet:::.tbl_element(tbl)
}
\keyword{internal}
